"""This module brings in-place mutations"""
from .base import BaseConversion, BaseMutation
from .heuristics import Weights
from .utils import Code


class BaseNameValueMutation(BaseMutation):
    """A base in-place mutation where name and value are needed to define it"""

    def __init__(self, name, value):
        """
        Args:
          name: to be wrapped with :py:obj:`ensure_conversion` and used as
            a key/attr/index for a mutation
          value: to be wrapped with :py:obj:`ensure_conversion` and used as
            a value for a mutation
        """
        super().__init__()
        self.name = self.ensure_conversion(name)
        self.value = self.ensure_conversion(value)


class BaseIndexMutation(BaseMutation):
    def __init__(self, index):
        super().__init__()
        self.index = self.ensure_conversion(index)


class SetItem(BaseNameValueMutation):
    def _gen_code_and_update_ctx(self, code_input, ctx):
        name_code = self.name.gen_code_and_update_ctx(code_input, ctx)
        value_code = self.value.gen_code_and_update_ctx(code_input, ctx)
        return f"{code_input}[{name_code}] = {value_code}"


class SetAttr(BaseNameValueMutation):
    def _gen_code_and_update_ctx(self, code_input, ctx):
        name_code = self.name.gen_code_and_update_ctx(code_input, ctx)
        value_code = self.value.gen_code_and_update_ctx(code_input, ctx)
        return f"setattr({code_input}, {name_code}, {value_code})"


class DelItem(BaseIndexMutation):
    def _gen_code_and_update_ctx(self, code_input, ctx):
        index_code = self.index.gen_code_and_update_ctx(code_input, ctx)
        return f"{code_input}.pop({index_code})"


class DelAttr(BaseIndexMutation):
    def _gen_code_and_update_ctx(self, code_input, ctx):
        index_code = self.index.gen_code_and_update_ctx(code_input, ctx)
        return f"delattr({code_input}, {index_code})"


class Custom(BaseMutation):
    """Mutation to be used in conjunction with `tap` method.
    Runs the code, defined by the conversion argument and returns the input
    as is."""

    def __init__(self, conversion):
        """
        Arg:
          conversion: to be wrapped with :py:obj:`ensure_conversion` and used
            as a mutation code
        """
        super().__init__()
        self.conversion = self.ensure_conversion(conversion)

    def _gen_code_and_update_ctx(self, code_input, ctx):
        return self.conversion.gen_code_and_update_ctx(code_input, ctx)


class Mutations:
    """Mutations to be used in conjunction with the conversion `tap` method."""

    #: Sets a value by key.
    set_item = SetItem
    #: Sets a value of an attribute.
    set_attr = SetAttr
    #: Pops a key from a dict.
    del_item = DelItem
    #: Deletes an attribute from an object.
    del_attr = DelAttr
    #: Runs the code, defined by the conversion argument and returns the input
    #: as is.
    custom = Custom


class TapConversion(BaseConversion):
    """This conversion generates the code which mutates the input data
    in-place.  TapConversion takes any number of mutations"""

    weight = Weights.FUNCTION_CALL

    def __init__(self, obj, *mutations: BaseMutation):
        super().__init__()
        self.obj = self.ensure_conversion(obj)
        self.mutations = [
            self.ensure_conversion(mut, explicitly_allowed_cls=BaseMutation)
            for mut in mutations
        ]
        self.number_of_input_uses = 1

    def _gen_code_and_update_ctx(self, code_input, ctx):
        suffix = self.gen_random_name("", ctx)
        converter_name = f"tap_{suffix}"
        function_ctx = self.as_function_ctx(ctx, optimize_naive=True)
        function_ctx.add_arg("data_", self.obj)
        with function_ctx:
            code = Code()
            code.add_line("def placeholder", 1)
            for mut in self.mutations:
                code.add_line(mut.gen_code_and_update_ctx("data_", ctx), 0)
            code.add_line("return data_", 0)

            code.lines_info[0] = (
                0,
                f"def {converter_name}({function_ctx.get_def_all_args_code()}):",
            )
            conversion = function_ctx.gen_conversion(
                converter_name, code.to_string(base_indent_level=0)
            )
        return function_ctx.call_with_all_args(
            conversion
        ).gen_code_and_update_ctx(code_input, ctx)


class IterMutConversion(TapConversion):
    """This conversion generates the code which iterates and mutates the
    elements in-place. The result is a generator.
    IterMutConversion takes any number of mutations"""

    def _gen_code_and_update_ctx(self, code_input, ctx):
        suffix = self.gen_random_name("", ctx)
        converter_name = f"iter_mut_{suffix}"
        code_item = f"item_{suffix}"
        function_ctx = self.as_function_ctx(ctx, optimize_naive=True)
        function_ctx.add_arg("data_", self.obj)

        with function_ctx:
            code = Code()
            code.add_line("def placeholder", 1)
            code.add_line(f"for {code_item} in data_:", 1)
            for mut in self.mutations:
                code.add_line(
                    mut.gen_code_and_update_ctx(f"{code_item}", ctx), 0
                )

            code.add_line(f"yield {code_item}", 0)

            code.lines_info[0] = (
                0,
                f"def {converter_name}({function_ctx.get_def_all_args_code()}):",
            )
            conversion = function_ctx.gen_conversion(
                converter_name, code.to_string(base_indent_level=0)
            )
        return function_ctx.call_with_all_args(
            conversion
        ).gen_code_and_update_ctx(code_input, ctx)



"""Provides conversions which simplify debug"""
import pdb
import sys

from .base import BaseConversion, ensure_conversion


if "pydevd" in sys.modules:  # pragma: no cover

    def debug_func(obj):
        import pydevd  # type: ignore # pylint: disable=import-outside-toplevel

        pydevd.settrace()
        return obj

elif sys.version_info[:2] < (3, 7):  # pragma: no cover

    def debug_func(obj):
        pdb.set_trace()  # pylint: disable=forgotten-debug-statement
        return obj

else:  # pragma: no cover

    def debug_func(obj):
        breakpoint()  # pylint: disable=undefined-variable,forgotten-debug-statement # noqa: F821,E501
        return obj


class Breakpoint(BaseConversion):
    """Defines the conversion which wraps another one and puts a breakpoint
    after it"""

    self_content_type = (
        BaseConversion.self_content_type
        | BaseConversion.ContentTypes.BREAKPOINT
    ) & ~BaseConversion.ContentTypes.FUNCTION_OF_INPUT
    debug_func = staticmethod(debug_func)

    def __init__(self, to_debug):
        super().__init__()
        self.conversion = self.ensure_conversion(
            ensure_conversion(to_debug).pipe(self.debug_func)
        )

    def _gen_code_and_update_ctx(self, code_input, ctx):
        return self.conversion.gen_code_and_update_ctx(code_input, ctx)



"""The only object is exported - conversion.
This object exposes the public API of conversions."""
from .conversion import conversion  # noqa: F401
from .dt import DateGrid, DateTimeGrid


__version__ = "0.42.3"



"""
Defines datetime utility functions
"""
from datetime import date, datetime, timedelta
from typing import Dict, Iterator, List, Type, Union


date_from_ordinal = date.fromordinal
datetime_from_ordinal = datetime.fromordinal
MICROSECOND = timedelta(microseconds=1)
DAY = timedelta(days=1)


__all__ = ["DateGrid", "DateTimeGrid"]


class TruncModes:
    """Defines truncate modes: START obviously means period start, END means
    the non-inclusive end of current interval (start of the next one),
    END_INCLUSIVE means inclusive end of the current interval"""

    START = 1
    END = 2
    END_INCLUSIVE = 3

    _verbose_to_value = {
        "start": START,
        "end": END,
        "end_inclusive": END_INCLUSIVE,
    }

    @classmethod
    def to_internal(cls, value):
        try:
            return cls._verbose_to_value[value]
        except KeyError:
            raise ValueError(  # pylint: disable=raise-missing-from
                "unsupported mode"
            )


class BaseStep:
    """Base definition of a time-related step"""

    types: "Dict[str, int]"

    def to_months(self):
        raise TypeError("cannot be interpreted as months")

    def to_days(self):
        raise TypeError("cannot be interpreted as days")

    def to_us(self):
        raise TypeError("cannot be interpreted as us")


class MonthStep(BaseStep):
    """Defines steps as number of months"""

    types = {"y": 12, "mo": 1}

    def __init__(self, params, negative=False):
        self.months = 0
        for type_, number in params.items():
            self.months += number * self.types[type_]
        if negative:
            self.months *= -1

    def to_months(self):
        return self.months


class DayOfWeekStep(BaseStep):
    """Defines steps as weeks, starting on certain day of week"""

    types = {
        "sun": 0,
        "mon": 1,
        "tue": 2,
        "wed": 3,
        "thu": 4,
        "fri": 5,
        "sat": 6,
    }

    def __init__(self, params):
        if len(params) != 1:
            raise AssertionError
        self.days = 7
        self.day_of_week_offset = 0
        for type_, number in params.items():
            self.days *= number
            self.day_of_week_offset += self.types[type_]

    def to_days(self):
        return self.days


class MicroSecondStep(BaseStep):
    """Defines steps as deterministic time units (days, hours, minutes,
    seconds, milliseconds, microseconds)"""

    types = {
        "d": 86400000000,
        "h": 3600000000,
        "m": 60000000,
        "s": 1000000,
        "ms": 1000,
        "us": 1,
    }

    def __init__(self, params, negative=False):
        self.us = 0
        for type_, number in params.items():
            self.us += number * self.types[type_]
        if negative:
            self.us *= -1

    def can_be_cast_to_days(self):
        return self.us % 86400000000 == 0

    def to_days(self):
        if not self.can_be_cast_to_days():
            raise TypeError("cannot be interpreted as whole days")
        return self.us // 86400000000

    def to_us(self):
        return self.us


def date_trunc_to_month(dt, to_months, offset_months, mode):
    total_months = dt.year * 12 + dt.month - 1
    total_months -= (total_months - offset_months) % to_months
    if mode == 1:
        return date(total_months // 12, total_months % 12 + 1, 1)

    elif mode == 2:
        total_months += to_months
        return date(total_months // 12, total_months % 12 + 1, 1)

    total_months += to_months
    return date(total_months // 12, total_months % 12 + 1, 1) - DAY


def date_trunc_to_day(dt, to_days, offset_days, mode):
    days = dt.toordinal()
    days -= (days - offset_days) % to_days
    if mode == 1:
        return date_from_ordinal(days)

    elif mode == 2:
        return date_from_ordinal(days + to_days)

    return date_from_ordinal(days + to_days - 1)


def datetime_trunc_to_month(dt, to_months, offset_months, mode):
    total_months = dt.year * 12 + dt.month - 1
    total_months -= (total_months - offset_months) % to_months
    if mode == 1:
        return datetime(
            total_months // 12, total_months % 12 + 1, 1, tzinfo=dt.tzinfo
        )

    elif mode == 2:
        total_months += to_months
        return datetime(
            total_months // 12, total_months % 12 + 1, 1, tzinfo=dt.tzinfo
        )

    total_months += to_months
    return (
        datetime(
            total_months // 12, total_months % 12 + 1, 1, tzinfo=dt.tzinfo
        )
        - MICROSECOND
    )


def datetime_trunc_to_day(dt, to_days, offset_days, mode):
    days = dt.toordinal()
    days -= (days - offset_days) % to_days
    if mode == 1:
        return datetime_from_ordinal(days).replace(tzinfo=dt.tzinfo)

    elif mode == 2:
        return datetime_from_ordinal(days + to_days).replace(tzinfo=dt.tzinfo)

    return (
        datetime_from_ordinal(days + to_days).replace(tzinfo=dt.tzinfo)
        - MICROSECOND
    )


def datetime_trunc_to_microsecond(dt, to_us, offset_us, mode):
    us = (
        (dt.toordinal() - 1) * 86400000000
        + dt.hour * 3600000000
        + dt.minute * 60000000
        + dt.second * 1000000
        + dt.microsecond
    )
    us -= (us - offset_us) % to_us

    if mode == 1:
        pass
    elif mode == 2:
        us += to_us
    else:
        us += to_us - 1

    new_dt = datetime_from_ordinal(us // 86400000000 + 1)
    left_microseconds = us % 86400000000
    return new_dt.replace(
        hour=left_microseconds // 3600000000,
        minute=left_microseconds % 3600000000 // 60000000,
        second=left_microseconds % 60000000 // 1000000,
        microsecond=left_microseconds % 1000000,
        tzinfo=dt.tzinfo,
    )


STEP_CLASSES: "List[Union[Type[MonthStep], Type[DayOfWeekStep], Type[MicroSecondStep]]]" = [
    MonthStep,
    DayOfWeekStep,
    MicroSecondStep,
]
type_to_cls = {type_: cls_ for cls_ in STEP_CLASSES for type_ in cls_.types}


def to_step(in_) -> "Union[MonthStep, DayOfWeekStep, MicroSecondStep]":
    if isinstance(in_, str):
        if not in_:
            raise ValueError("empty definition")

        length = len(in_)
        step_cls = None
        params = {}

        i = 0
        if in_.startswith("-"):
            negative = True
            i = 1
        else:
            negative = False

        while i < length:
            first_digit = i
            while i < length and in_[i].isdigit():
                i += 1

            if first_digit == i:
                number = 1
            else:
                number = int(in_[first_digit:i])

            first_symbol = i
            while i < length and not in_[i].isdigit():
                i += 1
            type_ = in_[first_symbol:i]

            if type_ in type_to_cls:
                if step_cls is None:
                    step_cls = type_to_cls[type_]
                elif type_to_cls[type_] is not step_cls:
                    raise ValueError(
                        "do not mix parameters of different steps",
                        step_cls.__name__,
                        type_to_cls[type_].__name__,
                    )

                if type_ in params:
                    raise ValueError(
                        f"step type is encountered twice: {type_}"
                    )

                params[type_] = number
            else:
                raise ValueError("unknown type", type_)

        if step_cls is None:
            raise AssertionError

        if negative:
            if issubclass(step_cls, DayOfWeekStep):
                raise ValueError("day-of-week steps cannot be negative")
            return step_cls(params, negative=True)

        return step_cls(params)

    elif isinstance(in_, timedelta):
        return MicroSecondStep({"us": round(in_.total_seconds() * 1000000)})

    raise ValueError("unsupported definition of grid")


def gen_dates__month(dt_start, dt_end, to_months, offset_months, mode):
    start_months = dt_start.year * 12 + dt_start.month - 1
    end_months = dt_end.year * 12 + dt_end.month - 1

    months = start_months - (start_months - offset_months) % to_months

    if mode == 1:
        while months <= end_months:
            yield date(months // 12, months % 12 + 1, 1)
            months += to_months

    elif mode == 2:
        while months <= end_months:
            months += to_months
            yield date(months // 12, months % 12 + 1, 1)

    else:
        while months <= end_months:
            months += to_months
            yield date(months // 12, months % 12 + 1, 1) - DAY


def gen_datetimes__month(dt_start, dt_end, to_months, offset_months, mode):
    start_months = dt_start.year * 12 + dt_start.month - 1
    end_months = dt_end.year * 12 + dt_end.month - 1
    tzinfo = dt_start.tzinfo

    months = start_months - (start_months - offset_months) % to_months

    if mode == 1:
        while months <= end_months:
            yield datetime(months // 12, months % 12 + 1, 1, tzinfo=tzinfo)
            months += to_months

    elif mode == 2:
        while months <= end_months:
            months += to_months
            yield datetime(months // 12, months % 12 + 1, 1, tzinfo=tzinfo)

    else:
        while months <= end_months:
            months += to_months
            yield datetime(
                months // 12, months % 12 + 1, 1, tzinfo=tzinfo
            ) - MICROSECOND


def gen_dates__day(dt_start, dt_end, to_days, offset_days, mode):
    start_days = dt_start.toordinal()
    end_days = dt_end.toordinal()

    days = start_days - (start_days - offset_days) % to_days

    if mode == 1:
        while days <= end_days:
            yield date_from_ordinal(days)
            days += to_days

    elif mode == 2:
        while days <= end_days:
            days += to_days
            yield date_from_ordinal(days)

    else:
        while days <= end_days:
            days += to_days
            yield date_from_ordinal(days - 1)


def gen_datetimes__day(dt_start, dt_end, to_days, offset_days, mode):
    start_days = dt_start.toordinal()
    end_days = dt_end.toordinal()
    tzinfo = dt_start.tzinfo

    days = start_days - (start_days - offset_days) % to_days

    if mode == 1:
        while days <= end_days:
            yield datetime_from_ordinal(days).replace(tzinfo=tzinfo)
            days += to_days

    elif mode == 2:
        while days <= end_days:
            days += to_days
            yield datetime_from_ordinal(days).replace(tzinfo=tzinfo)

    else:
        while days <= end_days:
            days += to_days
            yield datetime_from_ordinal(days).replace(
                tzinfo=tzinfo
            ) - MICROSECOND


def gen_datetimes__microsecond(dt_start, dt_end, to_us, offset_us, mode):
    tzinfo = dt_start.tzinfo
    start_us = (
        (dt_start.toordinal() - 1) * 86400000000
        + dt_start.hour * 3600000000
        + dt_start.minute * 60000000
        + dt_start.second * 1000000
        + dt_start.microsecond
    )
    end_us = (
        (dt_end.toordinal() - 1) * 86400000000
        + dt_end.hour * 3600000000
        + dt_end.minute * 60000000
        + dt_end.second * 1000000
        + dt_end.microsecond
    )

    us = start_us - (start_us - offset_us) % to_us

    if mode == 1:
        while us <= end_us:
            new_dt = datetime_from_ordinal(us // 86400000000 + 1)
            left_microseconds = us % 86400000000
            yield new_dt.replace(
                hour=left_microseconds // 3600000000,
                minute=left_microseconds % 3600000000 // 60000000,
                second=left_microseconds % 60000000 // 1000000,
                microsecond=left_microseconds % 1000000,
                tzinfo=tzinfo,
            )
            us += to_us

    elif mode == 2:
        while us <= end_us:
            us += to_us

            new_dt = datetime_from_ordinal(us // 86400000000 + 1)
            left_microseconds = us % 86400000000
            yield new_dt.replace(
                hour=left_microseconds // 3600000000,
                minute=left_microseconds % 3600000000 // 60000000,
                second=left_microseconds % 60000000 // 1000000,
                microsecond=left_microseconds % 1000000,
                tzinfo=tzinfo,
            )

    else:
        while us <= end_us:
            us += to_us

            new_dt = datetime_from_ordinal((us - 1) // 86400000000 + 1)
            left_microseconds = (us - 1) % 86400000000
            yield new_dt.replace(
                hour=left_microseconds // 3600000000,
                minute=left_microseconds % 3600000000 // 60000000,
                second=left_microseconds % 60000000 // 1000000,
                microsecond=left_microseconds % 1000000,
                tzinfo=tzinfo,
            )


class DateGrid:
    """Defines a grid of dates with a particular step and offset (e.g. every
    other Tuesday).

    Iterator of month grid, which contains the defined period:
    >>> it = DateGrid("mo").around(date(2020, 12, 31), date(2021, 1, 15))
    >>> assert list(it) == [date(2020, 12, 1), date(2021, 1, 1)]

    Same as above, but returns inclusive ends of periods
    >>> it = DateGrid("mo", mode="end_inclusive").around(
    >>>     date(2020, 12, 31), date(2021, 1, 15)
    >>> )
    >>> assert list(it) == [date(2020, 12, 31), date(2021, 1, 31)]

    Quarters:
    >>> DateGrid("3mo")

    Every other Tuesday:
    >>> DateGrid("2tue")

    Every 10 days grid, shifted 1 day forward:
    >>> DateGrid("10d", "1d")

    Both step and offset can be defined as a string which is comprised of
    numbers and suffixes:
     - y: year
     - mo: month
     - sun/mon/tue/wed/thu/fri/sat: days of week
     - d: day
     - h: hour
     - m: minute
     - s: second
     - ms: millisecond
     - us: microsecond

    so -2d8h10us means minus 2 days 8 hours and 10 microseconds.


    WARNING:
     * y/mo support only y/mo as offsets
     * days of week don't support offsets
     * as this method truncates dates, not datetimes, it accepts only whole
       number of days as steps and offsets

    Args:
      step: defines period length. If it's a year, month or a day of week,
        it defines beginnings of periods too.
      offset (optional): defines the shift of the expected date grid
        relative to 0-point. Positive offset shifts a grid to the right.
      mode (Literal["start", "end", "end_inclusive"]): defines truncating
        mode: "start" returns period start; "end" returns a start of the
        next period, complies with default interval definition where start
        is inclusive and end is not; "end_inclusive" returns the end of the
        current interval

    """

    __slots__ = ["f", "step", "offset", "mode"]

    def __init__(self, step, offset=None, mode="start"):
        self.mode = TruncModes.to_internal(mode)

        step = to_step(step)
        offset = None if offset is None else to_step(offset)

        if isinstance(step, MonthStep):
            self.f = gen_dates__month
            self.step = step.to_months()
            self.offset = 0 if offset is None else offset.to_months()

        elif isinstance(step, DayOfWeekStep):
            if offset is not None:
                raise ValueError(
                    "offsets are not applicable to day-of-week steps"
                )
            self.f = gen_dates__day
            self.step = step.to_days()
            self.offset = step.day_of_week_offset

        else:
            self.f = gen_dates__day
            self.step = step.to_days()
            self.offset = 0 if offset is None else offset.to_days()

    def around(
        self,
        dt_start: "Union[date, datetime]",
        dt_end: "Union[date, datetime]",
    ) -> "Iterator[date]":
        return self.f(dt_start, dt_end, self.step, self.offset, self.mode)


class DateTimeGrid:
    """Defines a grid of dates with a particular step and offset (e.g. every
    other Tuesday).

    Iterator of month grid, which contains the defined period:
    >>> it = DateTimeGrid("mo").around(
    >>>     datetime(2020, 12, 31),
    >>>     datetime(2021, 1, 15)
    >>> )
    >>> assert list(it) == [datetime(2020, 12, 1, 0, 0), datetime(2021, 1, 1, 0, 0)]

    Same as above, but returns inclusive ends of periods
    >>> it = DateTimeGrid("mo", mode="end_inclusive").around(
    >>>     datetime(2020, 12, 31), datetime(2021, 1, 15)
    >>> )
    >>> assert list(it) == [
    >>>     datetime(2020, 12, 31, 23, 59, 59, 999999),
    >>>     datetime(2021, 1, 31, 23, 59, 59, 999999)
    >>> ]

    Quarters:
    >>> DateTimeGrid("3mo")

    Every other Tuesday:
    >>> DateTimeGrid("2tue")

    Every 8 hours grid, shifted 6 hours forward:
    >>> DateTimeGrid("8h", "6h")

    STEP-STRING is a string which is comprised of numbers and suffixes:
     - y: year
     - mo: month
     - sun/mon/tue/wed/thu/fri/sat: days of week
     - d: day
     - h: hour
     - m: minute
     - s: second
     - ms: millisecond
     - us: microsecond

    so "-2d8h10us" means minus 2 days 8 hours and 10 microseconds.

    WARNING:
     * y/mo support only y/mo as offsets
     * days of week don't support offsets
     * any steps defined as deterministic units (d, h, m, s, ms, us) can
       only be used with offsets defined by deterministic units too

    Args:
      step: defines period length. If it's a year, month or a day of week,
        it defines beginnings of periods too.
      offset (optional): defines the shift of the expected date grid
        relative to 0-point. Positive offset shifts a grid to the right.
      mode (Literal["start", "end", "end_inclusive"]): defines truncating
        mode: "start" returns period start; "end" returns a start of the
        next period, complies with default interval definition where start
        is inclusive and end is not; "end_inclusive" returns the end of the
        current interval
    """

    __slots__ = ["f", "step", "offset", "mode"]

    def __init__(self, step, offset=None, mode="start"):
        self.mode = TruncModes.to_internal(mode)

        step = to_step(step)
        offset = None if offset is None else to_step(offset)

        if isinstance(step, MonthStep):
            self.f = gen_datetimes__month
            self.step = step.to_months()
            self.offset = 0 if offset is None else offset.to_months()

        elif isinstance(step, DayOfWeekStep):
            if offset is not None:
                raise ValueError(
                    "offsets are not applicable to day-of-week steps"
                )
            self.f = gen_datetimes__day
            self.step = step.to_days()
            self.offset = step.day_of_week_offset

        elif step.can_be_cast_to_days() and (
            offset is None or offset.can_be_cast_to_days()
        ):
            self.f = gen_datetimes__day
            self.step = step.to_days()
            self.offset = 0 if offset is None else offset.to_days()

        else:
            self.f = gen_datetimes__microsecond
            self.step = step.to_us()
            self.offset = 0 if offset is None else offset.to_us()

    def around(
        self, dt_start: datetime, dt_end: datetime
    ) -> "Iterator[datetime]":

        return self.f(dt_start, dt_end, self.step, self.offset, self.mode)


def date_parse(
    s,
    main_format,
    other_formats,
    strptime__=datetime.strptime,
    exc__=(ValueError, TypeError),
):
    try:
        return strptime__(s, main_format).date()
    except exc__:
        pass

    for fmt in other_formats:
        try:
            return strptime__(s, fmt).date()
        except exc__:
            pass

    raise ValueError("string doesn't match any format", s)


def datetime_parse(
    s,
    main_format,
    other_formats,
    strptime__=datetime.strptime,
    exc__=(ValueError, TypeError),
):
    try:
        return strptime__(s, main_format)
    except exc__:
        pass

    for fmt in other_formats:
        try:
            return strptime__(s, fmt)
        except exc__:
            pass

    raise ValueError("data doesn't match any format", s)



"""
This module brings join functionality to the library
"""
import typing as t
from itertools import chain, repeat

from .aggregations import Aggregate, ReduceFuncs
from .base import (
    And,
    BaseConversion,
    CallFunc,
    Eq,
    EscapedString,
    If,
    LazyEscapedString,
    ListComp,
    NaiveConversion,
    Namespace,
    This,
    Tuple,
)
from .columns import ColumnRef
from .utils import Code


_none = BaseConversion._none


class JoinException(Exception):
    pass


class JoinCondition(LazyEscapedString):
    NAME: str

    def __init__(self):
        super().__init__(self.NAME)

    def col(self, column_name: str) -> BaseConversion:
        return self.pipe(ColumnRef(column_name, id_=self.NAME))


class LeftJoinCondition(JoinCondition):
    NAME = "left_row"


class RightJoinCondition(JoinCondition):
    NAME = "right_row"


class _JoinConditions:
    """A helper object to analyze join conditions"""

    LEFT = LeftJoinCondition()
    RIGHT = RightJoinCondition()
    LEFT_NAME = LEFT.name
    RIGHT_NAME = RIGHT.name
    _ANY = {LEFT_NAME, RIGHT_NAME}

    def __init__(self, how="inner", swapped=False):
        if how == "right":
            raise AssertionError(
                "initialize via from_condition; it should have replaced right with left"
            )
        self.pre_filter = []
        self.inner_loop_conditions = []
        self.left_collection_filters = []
        self.left_row_hashers = []
        self.right_collection_filters = []
        self.right_row_hashers = []

        self.how = how
        self.swapped = swapped

        self.inner_join = how == "inner"
        self.left_join = how in {"left", "outer"}
        self.outer_join = how == "outer"
        if not self.inner_join and not self.left_join:
            raise AssertionError

    def _add_hashers(self, left, right):
        if self.swapped:
            self.right_row_hashers.append(left)
            self.left_row_hashers.append(right)
        else:
            self.left_row_hashers.append(left)
            self.right_row_hashers.append(right)

    def _add_left_filter(self, filter_conv, external_call=True):
        if self.swapped and external_call:
            return self._add_right_filter(filter_conv, external_call=False)

        if self.left_join:
            self.inner_loop_conditions.append(filter_conv)
        else:
            self.left_collection_filters.append(filter_conv)

    def _add_right_filter(self, filter_conv, external_call=True):
        if self.swapped and external_call:
            return self._add_left_filter(filter_conv, external_call=False)

        if self.outer_join:
            self.inner_loop_conditions.append(filter_conv)
        else:
            self.right_collection_filters.append(filter_conv)

    @classmethod
    def from_condition(cls, condition, how="inner") -> "_JoinConditions":
        if how == "right":
            join_conditions = cls(how="left", swapped=True)
        else:
            join_conditions = cls(how)

        if isinstance(condition, Namespace):
            condition = condition.conversion

        if isinstance(condition, NaiveConversion) and condition.value is True:
            return join_conditions

        if isinstance(condition, Eq):
            join_conditions.consume_eq(condition)
        elif isinstance(condition, And):
            join_conditions.consume_and(condition)
        else:
            join_conditions.consume_other(condition)

        return join_conditions

    @classmethod
    def _get_join_deps(cls, conv: "BaseConversion") -> t.Set[str]:
        return {
            dep.name
            for dep in conv.get_dependencies(types=LazyEscapedString)
            if dep.name in cls._ANY
        }

    def consume_eq(self, eq_conversion: "Eq"):
        if not isinstance(eq_conversion, Eq) or len(eq_conversion.args) <= 1:
            raise AssertionError

        if len(eq_conversion.args) > 2:
            for i in range(len(eq_conversion.args) - 1):
                self.consume_eq(
                    Eq(eq_conversion.args[i], eq_conversion.args[i + 1])
                )
            return

        first_deps, second_deps = map(self._get_join_deps, eq_conversion.args)
        if len(first_deps) > 1 or len(second_deps) > 1:
            return self.inner_loop_conditions.append(eq_conversion)

        if len(first_deps.union(second_deps)) == 2:
            if self.LEFT_NAME in first_deps:
                left_idx, right_idx = 0, 1
            else:
                left_idx, right_idx = 1, 0
            self._add_hashers(
                eq_conversion.args[left_idx], eq_conversion.args[right_idx]
            )
        else:
            all_args = first_deps.union(second_deps)
            if self.LEFT_NAME in all_args:
                self._add_left_filter(eq_conversion)
            elif self.RIGHT_NAME in all_args:
                self._add_right_filter(eq_conversion)
            else:
                self.pre_filter.append(eq_conversion)

    def consume_other(self, other: BaseConversion):
        deps = self._get_join_deps(other)
        deps_length = len(deps)
        if deps_length > 1:
            self.inner_loop_conditions.append(other)
        elif deps_length == 1:
            if self.LEFT_NAME in deps:
                self._add_left_filter(other)
            else:
                self._add_right_filter(other)
        else:
            self.pre_filter.append(other)

    def consume_and(self, and_conversion: And):
        if not isinstance(and_conversion, And):
            raise AssertionError

        for arg in and_conversion.args:
            if isinstance(arg, Eq):
                self.consume_eq(arg)
            elif isinstance(arg, And):
                self.consume_and(arg)
            else:
                self.consume_other(arg)

    def wrap_with_namespace(self, conversion, left=None, right=None):
        name_to_code = {}
        if left is not None:
            name_to_code[
                self.RIGHT_NAME if self.swapped else self.LEFT_NAME
            ] = left
        if right is not None:
            name_to_code[
                self.LEFT_NAME if self.swapped else self.RIGHT_NAME
            ] = right

        return Namespace(conversion, name_to_code)


class JoinConversion(BaseConversion):
    """Generates conversion which joins left_conversion and right_conversion
    using condition. The result is a generator of joined pairs

    Args:
      left_conversion (BaseConversion): left collection to join
      right_conversion (BaseConversion): right collection to join
      condition (BaseConversion): join condition. If is True, results in cross
        join
      how (str): one of the following: ``"inner"``, ``"left"``, ``"right"``,
        ``"outer"``
    """

    self_content_type = (
        BaseConversion.self_content_type
        | BaseConversion.ContentTypes.NONE_USAGE
    )

    def __init__(
        self,
        left_conversion: BaseConversion,
        right_conversion: BaseConversion,
        condition: BaseConversion,
        how="inner",
    ):
        super().__init__()
        self.left_conversion = self.ensure_conversion(left_conversion)
        self.right_conversion = self.ensure_conversion(right_conversion)

        # hiding left & right LazyEscapedString from parents not to affect
        # parent function args
        self.condition = self.ensure_conversion(
            Namespace(
                condition,
                name_to_code={
                    _JoinConditions.LEFT_NAME: None,
                    _JoinConditions.RIGHT_NAME: None,
                },
            )
        )
        self.how = self.validate_how(how)

    @classmethod
    def validate_how(cls, how: str):
        how = how.lower()
        if how not in ("inner", "left", "right", "outer", "full"):
            raise ValueError(how)
        if how == "full":
            how = "outer"
        return how

    def _gen_code_and_update_ctx(self, code_input, ctx):
        join_conditions = _JoinConditions.from_condition(
            self.condition, how=self.how
        )

        suffix = self.gen_random_name("", ctx)
        converter_name = f"join{suffix}"
        code = Code()
        function_ctx = self.condition.as_function_ctx(ctx, optimize_naive=True)

        if join_conditions.swapped:
            function_ctx.add_arg("left_", self.right_conversion)
            function_ctx.add_arg("right_", self.left_conversion)
        else:
            function_ctx.add_arg("left_", self.left_conversion)
            function_ctx.add_arg("right_", self.right_conversion)

        function_ctx.add_arg("_none", EscapedString("_none"))

        with function_ctx:
            code.add_line("def placeholder", 1)

            if join_conditions.left_collection_filters:
                code.add_line(
                    "left_ = %s"
                    % This.filter(
                        join_conditions.wrap_with_namespace(
                            And(*join_conditions.left_collection_filters),
                            left=True,
                        )
                    ).gen_code_and_update_ctx("left_", ctx),
                    0,
                )

            if join_conditions.outer_join:
                code.add_line("yielded_right_ids = set()", 0)

            def track_id():
                if join_conditions.outer_join:
                    code.add_line("yielded_right_ids.add(id(right_item))", 0)

            if join_conditions.right_row_hashers:
                c_left_key_to_hash = join_conditions.wrap_with_namespace(
                    Tuple(*join_conditions.left_row_hashers)
                    if len(join_conditions.left_row_hashers) > 1
                    else join_conditions.left_row_hashers[0],
                    left=True,
                )
                c_right_key_to_hash = join_conditions.wrap_with_namespace(
                    Tuple(*join_conditions.right_row_hashers)
                    if len(join_conditions.right_row_hashers) > 1
                    else join_conditions.right_row_hashers[0],
                    right=True,
                )
                code.add_line(
                    "hash_to_right_items = %s"
                    % EscapedString("right_")
                    .pipe(
                        Aggregate(
                            ReduceFuncs.DictArray(
                                c_right_key_to_hash,
                                This,
                                default=NaiveConversion({}),
                                where=join_conditions.wrap_with_namespace(
                                    And(
                                        *join_conditions.right_collection_filters
                                    ),
                                    right=True,
                                )
                                if join_conditions.right_collection_filters
                                else None,
                            )
                        )
                    )
                    .gen_code_and_update_ctx("right_", ctx),
                    0,
                )
                code.add_line("del right_", 0)
                initial_right = "(item for items in hash_to_right_items.values() for item in items)"

                code.add_line("for left_item in left_:", 1)
                code.add_line(
                    "left_key = %s"
                    % c_left_key_to_hash.gen_code_and_update_ctx(
                        "left_item", ctx
                    ),
                    0,
                )

                c_left_key = EscapedString("left_key")
                c_hash_to_right_items = EscapedString("hash_to_right_items")

                code.add_line(
                    "right_items = %s"
                    % If(
                        c_left_key.in_(c_hash_to_right_items),
                        c_hash_to_right_items.item(c_left_key).pipe(
                            This.filter(
                                join_conditions.wrap_with_namespace(
                                    And(
                                        *join_conditions.inner_loop_conditions
                                    ),
                                    left="left_item",
                                    right=True,
                                )
                            )
                            if join_conditions.inner_loop_conditions
                            else This
                        ),
                        Tuple(),
                    )
                    .pipe(iter if join_conditions.left_join else This)
                    .gen_code_and_update_ctx(None, ctx),
                    0,
                )

            else:

                if join_conditions.right_collection_filters:
                    code.add_line(
                        "right_ = %s"
                        % ListComp(
                            This,
                            join_conditions.wrap_with_namespace(
                                And(*join_conditions.right_collection_filters),
                                right=True,
                            ),
                            _none,
                        ).gen_code_and_update_ctx("right_", ctx),
                        0,
                    )
                else:
                    code.add_line(
                        "right_ = %s"
                        % If(
                            CallFunc(isinstance, This, t.Sized),
                            This,
                            This.pipe(list),
                        ).gen_code_and_update_ctx("right_", ctx),
                        0,
                    )
                initial_right = "right_"

                code.add_line("for left_item in left_:", 1)
                code.add_line(
                    "right_items = %s"
                    % (
                        This.filter(
                            Namespace(
                                And(*join_conditions.inner_loop_conditions),
                                name_to_code={
                                    _JoinConditions.LEFT_NAME: "left_item",
                                    _JoinConditions.RIGHT_NAME: True,
                                },
                            )
                        )
                        if join_conditions.inner_loop_conditions
                        else This
                    )
                    .pipe(iter if join_conditions.left_join else This)
                    .gen_code_and_update_ctx("right_", ctx),
                    0,
                )

            if join_conditions.left_join:
                code.add_line("right_item = next(right_items, _none)", 0)
                code.add_line("if right_item is _none:", 1)
                code.add_line(
                    "yield None, left_item"
                    if join_conditions.swapped
                    else "yield left_item, None",
                    -1,
                )
                code.add_line("else:", 1)
                track_id()
                code.add_line(
                    "yield right_item, left_item"
                    if join_conditions.swapped
                    else "yield left_item, right_item",
                    0,
                )
                code.add_line(
                    "for right_item in right_items:",
                    1,
                )
                track_id()
                code.add_line(
                    "yield right_item, left_item"
                    if join_conditions.swapped
                    else "yield left_item, right_item",
                    -3,
                )

            else:
                code.add_line("for right_item in right_items:", 1)
                track_id()
                code.add_line(
                    "yield right_item, left_item"
                    if join_conditions.swapped
                    else "yield left_item, right_item",
                    -2,
                )

            if join_conditions.outer_join:
                code.add_line(
                    "yield from ("
                    "(None, right_item) "
                    f"for right_item in {initial_right} "
                    "if id(right_item) not in yielded_right_ids)",
                    0,
                )

            code.lines_info[0] = (
                0,
                f"def {converter_name}({function_ctx.get_def_all_args_code()}):",
            )
            conversion = function_ctx.gen_conversion(
                converter_name, code.to_string(0)
            )
        c_result = function_ctx.call_with_all_args(conversion)

        if join_conditions.pre_filter:
            if join_conditions.inner_join:
                c_result = If(
                    And(*join_conditions.pre_filter),
                    c_result,
                    Tuple().pipe(iter),
                )
            elif join_conditions.outer_join:
                c_result = If(
                    And(*join_conditions.pre_filter),
                    c_result,
                    CallFunc(
                        chain,
                        CallFunc(zip, self.left_conversion, repeat(None)),
                        CallFunc(zip, repeat(None), self.right_conversion),
                    ),
                )
            else:
                c_result = If(
                    And(*join_conditions.pre_filter),
                    c_result,
                    CallFunc(zip, repeat(None), self.right_conversion)
                    if join_conditions.swapped
                    else CallFunc(zip, self.left_conversion, repeat(None)),
                )

        return c_result.gen_code_and_update_ctx(code_input, ctx)



"""
Python's native open() doesn't support custom newlines in the text mode and
doesn't support "newlines" (delimiters) in binary mode. The following methods
should close the gap.
"""


def split_buffer(buffer, delimiter, chunk_size=32768):
    """Reads text or binary buffer and splits it by delimiter.

    Args:
      buffer: buffer to be read
      delimiter: delimiter to use for splitting
      chunk_size: chunk size to read at every iteration
    """
    delimiter_length = len(delimiter)
    with buffer:
        chunk = buffer.read(chunk_size)
        if not chunk:
            yield chunk
            return
        checked_length = 0
        while True:
            new_chunk = buffer.read(chunk_size)
            chunk = chunk + new_chunk

            if chunk.find(delimiter, checked_length) != -1:
                chunks = chunk.split(delimiter)
                for i in range(len(chunks) - 1):
                    yield chunks[i]
                chunk = chunks[-1]
                del chunks

            chunk_length = len(chunk)
            checked_length = (
                (chunk_length - delimiter_length)
                if chunk_length > delimiter_length
                else 0
            )

            if not new_chunk:
                yield chunk
                break


def split_buffer_n_decode(
    buffer, delimiter, chunk_size=32768, encoding="utf-8"
):
    """Reads binary buffer, splits it by binary delimiter and yields decoded
    chunks.

    Args:
      buffer: buffer to be read
      delimiter: delimiter to use for splitting
      chunk_size: chunk size to read at every iteration
      encoding: encoding to use when decoding a chunk
    """
    delimiter_length = len(delimiter)
    with buffer:
        chunk = buffer.read(chunk_size)
        if not chunk:
            yield chunk.decode(encoding)
            return
        checked_length = 0
        while True:
            new_chunk = buffer.read(chunk_size)
            chunk = chunk + new_chunk

            if chunk.find(delimiter, checked_length) != -1:
                chunks = chunk.split(delimiter)
                for i in range(len(chunks) - 1):
                    yield chunks[i].decode(encoding)
                chunk = chunks[-1]
                del chunks

            chunk_length = len(chunk)
            checked_length = (
                (chunk_length - delimiter_length)
                if chunk_length > delimiter_length
                else 0
            )

            if not new_chunk:
                yield chunk.decode(encoding)
                break



"""Provides conversions for working with unique values"""
from .base import BaseConversion, EscapedString
from .utils import Code


class IterUnique(BaseConversion):
    """
    Defines a conversion which iterates over the input and yields unique ones
    (supports custom uniqueness conditions)
    """

    def __init__(self, self_conv, element, by_):
        super().__init__()
        self.self_conv = self.ensure_conversion(self_conv)
        self.element = self.ensure_conversion(element)
        self.by_ = self.ensure_conversion(by_)

        self.input_arg_container = EscapedString("")
        self.input_arg_container.depends_on(self.element)
        self.input_arg_container.depends_on(self.by_)

    def _gen_code_and_update_ctx(self, code_input, ctx):
        converter_name = self.gen_random_name("iter_unique", ctx)
        function_ctx = self.input_arg_container.as_function_ctx(
            ctx, optimize_naive=True
        )
        function_ctx.add_arg("data_", self.self_conv)

        code = Code()
        code.add_line("def placeholder", 1)

        with function_ctx:
            code.add_line("s_ = set()", 0)
            code.add_line("s_add = s_.add", 0)
            code.add_line("for item_ in data_:", 1)

            item_code = self.element.gen_code_and_update_ctx("item_", ctx)
            by_code = self.by_.gen_code_and_update_ctx("item_", ctx)
            if by_code != "item_":
                code.add_line(f"by_ = {by_code}", 0)
                code.add_line("if by_ not in s_:", 1)
                code.add_line("s_add(by_)", 0)
                if by_code == item_code:
                    code.add_line("yield by_", 0)
                else:
                    code.add_line(f"yield {item_code}", 0)
            else:
                code.add_line("if item_ not in s_:", 1)
                code.add_line("s_add(item_)", 0)
                code.add_line(f"yield {item_code}", 0)

            code.lines_info[0] = (
                0,
                f"def {converter_name}({function_ctx.get_def_all_args_code()}):",
            )
            conversion = function_ctx.gen_conversion(
                converter_name, code.to_string(base_indent_level=0)
            )
        return function_ctx.call_with_all_args(
            conversion
        ).gen_code_and_update_ctx(code_input, ctx)



"""This module brings aggregations with various reduce functions"""
import typing as t
from collections import defaultdict
from decimal import Decimal
from math import ceil

from .base import (
    BaseConversion,
    CallFunc,
    ConversionException,
    ConverterOptionsCtx,
    EscapedString,
    GeneratorItem,
    GetItem,
    If,
    InlineExpr,
    LazyEscapedString,
    List,
    ListComp,
    NaiveConversion,
    Namespace,
    NamespaceCtx,
    This,
    Tuple,
    _None,
    ensure_conversion,
)
from .heuristics import Weights
from .utils import Code


_none = BaseConversion._none
RBT = t.TypeVar("RBT", bound="ReduceBlock")


class ReduceBlock:
    """Represents a section of code of a single reducer"""

    var_checksum = "checksum_"

    def __init__(
        self,
        reduce_initial: t.Iterable[str],
        reduce_two: t.Iterable[str],
        reducer_code_input: str,
        var_agg_data_value: str,
        checksum_flag: int,
        unconditional_init: bool,
    ):
        self.reduce_initial = {var_agg_data_value: reduce_initial}
        self.reduce_two = {var_agg_data_value: reduce_two}
        self.reducer_code_input = reducer_code_input
        self.condition_code = None
        self.unconditional_init = unconditional_init
        self.checksum_flag = checksum_flag

    def consume(self: RBT, reduce_block: RBT) -> RBT:
        if (
            self.condition_code != reduce_block.condition_code
            or not self.unconditional_init
            or not reduce_block.unconditional_init
        ):
            raise AssertionError

        for k, v in reduce_block.reduce_initial.items():
            if k in self.reduce_initial:
                raise AssertionError
            self.reduce_initial[k] = v

        for k, v in reduce_block.reduce_two.items():
            if k in self.reduce_two:
                raise AssertionError
            self.reduce_two[k] = v

        self.checksum_flag += reduce_block.checksum_flag
        return self

    def iter_reduce_lines(self, lines_info_dict) -> t.Iterable[str]:
        for var_agg_data_value, lines in lines_info_dict.items():
            template_kwargs = {
                "result": var_agg_data_value,
                "input": self.reducer_code_input,
            }
            yield from (line % template_kwargs for line in lines)

    def iter_reduce_initial_lines(self) -> t.Iterable[str]:
        yield from self.iter_reduce_lines(self.reduce_initial)

    def iter_reduce_two_lines(self) -> t.Iterable[str]:
        yield from self.iter_reduce_lines(self.reduce_two)

    def as_key(self):
        return (
            self.condition_code,
            self.unconditional_init,
            self.reducer_code_input,
            tuple(
                line
                for lines in self.reduce_initial.values()
                for line in lines
            ),
            tuple(
                line for lines in self.reduce_two.values() for line in lines
            ),
        )


class ReduceConditionalBlock(ReduceBlock):
    """Represents a section of code of a single reducer with an incoming
    condition"""

    def __init__(self, *args, **kwargs):
        condition_code = kwargs.pop("condition_code")
        super().__init__(*args, **kwargs)
        self.condition_code = condition_code


ReduceBlockType = t.Union[ReduceBlock, ReduceConditionalBlock]


class ReduceBlocks:
    """Represents a set of reduce blocks"""

    def __init__(self, var_agg_data, aggregate_mode):
        self.var_agg_data = var_agg_data
        self.aggregate_mode = aggregate_mode

        self.block_key_to_index = {}
        self.agg_data_values = []

        self.condition_to_blocks = {}
        self.other_blocks = []
        self.number = 0

    def exchange_reducer_code_input(self, reducer, reducer_code_input, ctx):
        checksum_flag = 1 if self.aggregate_mode else 0
        agg_data_value = self._gen_agg_data_value(self.number)
        reduce_block = reducer.gen_reduce_code_block(
            agg_data_value,
            reducer_code_input,
            checksum_flag,
            ctx,
        )
        key = reduce_block.as_key()
        index = self.block_key_to_index.get(key)

        if index is not None:
            return self._gen_agg_data_value(index)

        self.block_key_to_index[key] = self.number
        self._add_block(reduce_block)
        self.agg_data_values.append(agg_data_value)
        return agg_data_value

    def _gen_agg_data_value(self, index):
        return (
            f"{self.var_agg_data}_v{index}"
            if self.aggregate_mode
            else f"{self.var_agg_data}.v{index}"
        )

    def _add_block(self, reduce_block: ReduceBlockType):
        self.number += 1

        if reduce_block.unconditional_init:
            condition = (
                reduce_block.condition_code
                if isinstance(reduce_block, ReduceConditionalBlock)
                else None
            )

            if condition in self.condition_to_blocks:
                self.condition_to_blocks[condition].consume(reduce_block)
            else:
                self.condition_to_blocks[condition] = reduce_block

        else:
            return self.other_blocks.append(reduce_block)

    def iter_blocks(self) -> t.Generator[ReduceBlockType, None, None]:
        yield from self.condition_to_blocks.values()
        yield from self.other_blocks

    def gen_group_by_code(
        self,
        var_row,
        var_agg_data,
        var_signature_to_agg_data,
        code_signature,
    ) -> Code:
        code = Code()
        code.add_line(f"for {var_row} in data_:", 1)
        code.add_line(
            f"{var_agg_data} = {var_signature_to_agg_data}[{code_signature}]",
            0,
        )

        for block in self.iter_blocks():
            reduce_initial_lines = list(block.iter_reduce_initial_lines())
            reduce_two_lines = list(block.iter_reduce_two_lines())
            any_var_agg_data_value = next(iter(block.reduce_initial))

            if block.condition_code:
                code.add_line(f"if {block.condition_code}:", 1)

            code.add_line(f"if {any_var_agg_data_value} is _none:", 1)
            for line in reduce_initial_lines:
                code.add_line(line, 0)
            code.incr_indent_level(-1)

            if reduce_two_lines:
                code.add_line("else:", 1)
                for line in reduce_two_lines:
                    code.add_line(line, 0)
                code.incr_indent_level(-1)

            if block.condition_code:
                code.incr_indent_level(-1)
        return code

    def gen_aggregate_code(
        self,
        var_row,
        expected_checksum,
    ) -> Code:
        is_debug = ConverterOptionsCtx.get_option_value("debug")
        var_checksum = ReduceBlock.var_checksum
        first_phase_code = Code()
        second_phase_code = Code()

        needs_sum_checking = False

        reduce_blocks = list(self.iter_blocks())
        is_single_block = len(reduce_blocks) == 1
        for block in reduce_blocks:
            reduce_initial_lines = list(block.iter_reduce_initial_lines())
            reduce_two_lines = list(block.iter_reduce_two_lines())
            any_var_agg_data_value = next(iter(block.reduce_initial))

            init_phase_needs_else = not (
                is_single_block and block.unconditional_init
            )

            if block.condition_code:
                first_phase_code.add_line(f"if {block.condition_code}:", 1)

            first_phase_code.add_line(
                f"if {any_var_agg_data_value} is _none:", 1
            )
            for line in reduce_initial_lines:
                first_phase_code.add_line(line, 0)

            if is_single_block:
                if block.unconditional_init:
                    if is_debug:
                        first_phase_code.add_line(
                            "globals()['__BROKEN_EARLY__'] = True  # DEBUG ONLY",
                            0,
                        )
                    first_phase_code.add_line("break", 0)
                else:
                    first_phase_code.add_line(
                        f"if {any_var_agg_data_value} is not _none:", 1
                    )
                    if is_debug:
                        first_phase_code.add_line(
                            "globals()['__BROKEN_EARLY__'] = True  # DEBUG ONLY",
                            0,
                        )
                    first_phase_code.add_line("break", -1)

            else:
                if block.unconditional_init:
                    first_phase_code.add_line(
                        f"{var_checksum} += {block.checksum_flag}", 0
                    )
                else:
                    first_phase_code.add_line(
                        f"if {any_var_agg_data_value} is not _none:", 1
                    )
                    first_phase_code.add_line(
                        f"{var_checksum} += {block.checksum_flag}", -1
                    )
                needs_sum_checking = True

            first_phase_code.incr_indent_level(-1)

            if init_phase_needs_else and reduce_two_lines:
                first_phase_code.add_line("else:", 1)
                for line in reduce_two_lines:
                    first_phase_code.add_line(line, 0)
                first_phase_code.incr_indent_level(-1)

            if block.condition_code:
                first_phase_code.incr_indent_level(-1)

            if reduce_two_lines:
                if block.condition_code:
                    second_phase_code.add_line(
                        f"if {block.condition_code}:", 1
                    )

                for line in reduce_two_lines:
                    second_phase_code.add_line(line, 0)

                if block.condition_code:
                    second_phase_code.incr_indent_level(-1)

        resulting_code = Code()

        resulting_code.add_line("it_ = iter(data_)", 0)
        resulting_code.add_line(f"for {var_row} in it_:", 1)

        resulting_code.add_code(first_phase_code)
        if needs_sum_checking:
            resulting_code.add_line(
                f"if {var_checksum} == {expected_checksum}:", 1
            )
            if is_debug:
                resulting_code.add_line(
                    "globals()['__BROKEN_EARLY__'] = True  # DEBUG ONLY", 0
                )
            resulting_code.add_line("break", -1)
        resulting_code.incr_indent_level(-1)

        if second_phase_code.has_lines():
            resulting_code.add_line("", 0)
            resulting_code.add_line(f"for {var_row} in it_:", 1)

            resulting_code.add_code(second_phase_code)
            resulting_code.incr_indent_level(-1)
        return resulting_code


GROUPER_TEMPLATE = """
def {converter_name}({code_args}):
    {var_signature_to_agg_data} = defaultdict({var_agg_data_cls})

{code_group_by}

{code_result}
"""
AGGREGATE_TEMPLATE = """
def {converter_name}({code_args}):
    {code_init_agg_vars}
    {var_checksum} = 0

{code_aggregate}

{code_result}
"""


RT = t.TypeVar("RT", bound="BaseReducer")


class BaseReducer(BaseConversion):
    """Base of a reduce operation to be used during the aggregation"""

    expressions: t.Tuple[t.Any, ...]
    post_conversion: t.Optional[BaseConversion] = None
    default: t.Any
    where: t.Any = _none
    unconditional_init: bool = False

    self_content_type = (
        (
            BaseConversion.self_content_type
            & ~BaseConversion.ContentTypes.FUNCTION_OF_INPUT
        )
        | BaseConversion.ContentTypes.REDUCER
        | BaseConversion.ContentTypes.NONE_USAGE
    )

    def __init__(self, default, initial):
        super().__init__()

        self.default, self.initial = self.prepare_default_n_initial(
            default, initial
        )
        self.conversion = self.ensure_conversion(
            If(
                This().is_(EscapedString("_none")),
                self.default,
                (
                    self.post_conversion
                    if self.post_conversion is not None
                    else This()
                ),
            )
        )

    def gen_reduce_code_block(
        self,
        var_agg_data_value: str,
        reducer_code_input: str,
        checksum_flag: int,
        ctx: dict,
    ) -> ReduceBlockType:
        raise NotImplementedError

    def _gen_code_and_update_ctx(self, code_input, ctx) -> str:
        code_input = ctx["exchange_reducer_code_input_methods"][-1](
            reducer=self,
            reducer_code_input=code_input,
            ctx=ctx,
        )
        return self.conversion.gen_code_and_update_ctx(code_input, ctx)

    def prepare_default_n_initial(self, default, initial):
        if default is _none:
            try:
                default = self.default
            except AttributeError:
                pass

        if initial is _none and hasattr(self, "initial"):
            initial = self.initial
        if initial is not _none:
            initial = self.ensure_conversion(initial)
            if initial.is_itself_callable_like():
                initial = initial.call_like()

            if default is _none and initial.ignores_input():
                default = initial

        if default is _none:
            raise ValueError("default is not provided")

        default = self.ensure_conversion(default)
        if default.is_itself_callable_like():
            default = default.call_like()

        return default, initial


class MultiStatementReducer(BaseReducer):
    """Defines the reduce operation, which is based on multiple python
    statements, to be used during the aggregation"""

    base_condition_code: "t.Optional[str]" = None
    prepare_first: t.Tuple[str, ...]
    reduce: t.Tuple[str, ...]

    def __init__(self, *expressions, initial=_none, default=_none, where=None):
        super().__init__(default, initial)
        self.where = (
            _none
            if (where is None or where is _none)
            else self.ensure_conversion(where)
        )

        self.expressions = tuple(
            self.ensure_conversion(expr)
            for expr in (
                self.expressions
                if hasattr(self, "expressions") and not expressions
                else expressions
            )
        )

    def _format_statements(
        self,
        statements,
        args,
        reducer_code_input,
        ctx,
        prev_result="%(result)s",
    ):
        code_args = tuple(
            arg.gen_code_and_update_ctx(reducer_code_input, ctx)
            for arg in args
        )
        return [
            statement.format(*code_args, prev_result=prev_result)
            for statement in statements
        ]

    def gen_reduce_code_block(
        self,
        var_agg_data_value: str,
        reducer_code_input: str,
        checksum_flag: int,
        ctx: dict,
    ) -> ReduceBlockType:
        if self.initial is _none:
            reduce_initial = self._format_statements(
                (
                    self.prepare_first(ctx)
                    if callable(self.prepare_first)
                    else self.prepare_first
                ),
                self.expressions,
                reducer_code_input,
                ctx,
            )
        else:
            reduce_initial = self._format_statements(
                (self.reduce(ctx) if callable(self.reduce) else self.reduce),
                self.expressions,
                reducer_code_input,
                ctx,
                prev_result=self.initial.gen_code_and_update_ctx(
                    reducer_code_input, ctx
                ),
            )

        reduce_two = self._format_statements(
            (self.reduce(ctx) if callable(self.reduce) else self.reduce),
            self.expressions,
            reducer_code_input,
            ctx,
        )

        kwargs = {
            "reduce_initial": reduce_initial,
            "reduce_two": reduce_two,
            "reducer_code_input": reducer_code_input,
            "var_agg_data_value": var_agg_data_value,
            "checksum_flag": checksum_flag,
            "unconditional_init": self.unconditional_init,
        }
        if self.where is not _none:
            condition_code = self.where.gen_code_and_update_ctx(
                reducer_code_input, ctx
            )
            if condition_code != "True":
                kwargs["condition_code"] = condition_code

        if self.base_condition_code:
            base_condition_code = self._format_statements(
                [self.base_condition_code],
                self.expressions,
                reducer_code_input,
                ctx,
            )[0]
            if "condition_code" in kwargs:
                kwargs[
                    "condition_code"
                ] = f"{base_condition_code} and {kwargs['condition_code']}"
            else:
                kwargs["condition_code"] = base_condition_code

        block_cls = (
            ReduceConditionalBlock
            if "condition_code" in kwargs
            else ReduceBlock
        )
        return block_cls(**kwargs)


class Reduce(BaseReducer):
    """Defines the reduce operation, which is based on a callable / expression
    to be used during the aggregation"""

    initial: t.Any

    def __init__(
        self,
        to_call_with_2_args: t.Union[t.Callable, InlineExpr],
        *expressions: t.Tuple[t.Any, ...],
        initial: t.Union[_None, t.Callable, InlineExpr, t.Any],
        default: t.Union[_None, t.Callable, t.Any] = _none,
        unconditional_init: bool = False,
        where=None,
    ):
        """
        Args:
          to_call_with_2_args: defines the reduce function/expression
          expressions: args to be passed to `to_call_with_2_args` after the
            aggregation value
          initial: defines the very first item to be passed to
            `to_call_with_2_args` item. If callable, then the result of a call
            is used. If a conversion is passed, it is resolved on the first
            row met.
          default: defines the value to be returned when there was nothing to
            reduce in a group (e.g. the current reduce operation has filtered
            out some rows, while an adjacent reduce operation has got
            something to reduce, forming a group). If callable, then the result
            of a call is used.  When default is not passed, initial is used if
            it doesn't depend on input data.
          unconditional_init: tells whether the first call initializes the
            aggregation value OR there is a condition for that
        """
        super().__init__(default, initial)
        self.where = (
            _none
            if (where is None or where is _none)
            else self.ensure_conversion(where)
        )
        self.to_call_with_2_args = self.ensure_conversion(to_call_with_2_args)
        self.expressions = tuple(
            self.ensure_conversion(expr) for expr in expressions
        )
        self.unconditional_init = unconditional_init

    def gen_reduce_code_block(
        self,
        var_agg_data_value: str,
        reducer_code_input: str,
        checksum_flag: int,
        ctx: dict,
    ) -> ReduceBlockType:
        _ = self.to_call_with_2_args.call_like(
            self.initial,
            *self.expressions,
        ).gen_code_and_update_ctx(reducer_code_input, ctx)
        reduce_initial = [f"%(result)s = {_}"]
        _ = self.to_call_with_2_args.call_like(
            EscapedString("%(result)s"),
            *self.expressions,
        ).gen_code_and_update_ctx(reducer_code_input, ctx)
        reduce_two = [f"%(result)s = {_}"]

        kwargs = {
            "reduce_initial": reduce_initial,
            "reduce_two": reduce_two,
            "reducer_code_input": reducer_code_input,
            "var_agg_data_value": var_agg_data_value,
            "checksum_flag": checksum_flag,
            "unconditional_init": self.unconditional_init,
        }
        if self.where is not _none:
            condition_code = self.where.gen_code_and_update_ctx(
                reducer_code_input, ctx
            )
            if condition_code != "True":
                kwargs["condition_code"] = condition_code

        block_cls = (
            ReduceConditionalBlock
            if "condition_code" in kwargs
            else ReduceBlock
        )

        return block_cls(**kwargs)


class GroupBy:
    """Generates the function which aggregates the data, grouping by
    conversions, specified in `__init__` method and returns list of items in a
    format defined by the parameter passed to ``aggregate`` method.

    If no group keys are passed, then it returns just a single value, defined
    by the parameter passed to ``aggregate`` method.

    Current optimizations:
     * piping like ``c.group_by(...).aggregate().pipe(...)`` won't run
       the aggregation twice, this is handled as 2 statements
     * using the same reduce clause twice (e.g. one used as an argument
       for some function calls) won't result in calculating this reduce twice
    """

    def __init__(self, *by):
        """Takes any number of conversions to group by

        Args:
          by (tuple): each item is to be wrapped with
            :py:obj:`ensure_conversion`.  Each is to resolve to a hashable
            object to allow using such tuples as keys. If nothing is passed,
            aggregate the input into a single object.
        """
        self.by = by

    def aggregate(
        self, reducer: t.Union[dict, list, set, tuple, BaseConversion]
    ) -> "Grouper":

        return Grouper(self.by, reducer)


def delegate_input_switching_method(name, force_iter_first=False):
    def method(self, *args, **kwargs):
        conversion = self.conversion
        if force_iter_first:
            conversion = conversion.to_iter()
        return Grouper(
            by=self.by,
            reducer=self.reducer,
            conversion=getattr(conversion, name)(*args, **kwargs),
        )

    return method


class Grouper(BaseConversion):
    """Fully initialized GroupBy conversion, which delegates some of methods
    like iter to its internals"""

    self_content_type = (
        BaseConversion.self_content_type
        | BaseConversion.ContentTypes.AGGREGATION
        | BaseConversion.ContentTypes.NONE_USAGE
    )

    SIGNATURE_NAME = "signature"
    AGG_DATA_NAME = "agg_data"
    AGG_RESULT_ITEM_NAME = "agg_result_item"
    SIGNATURE = Namespace(
        LazyEscapedString(SIGNATURE_NAME), {SIGNATURE_NAME: None}
    )
    AGG_DATA = Namespace(
        LazyEscapedString(AGG_DATA_NAME), {AGG_DATA_NAME: None}
    )
    AGG_RESULT_ITEM = Namespace(
        LazyEscapedString(AGG_RESULT_ITEM_NAME), {AGG_RESULT_ITEM_NAME: None}
    )
    AGG_RESULT_ITEM.weight = Weights.UNPREDICTABLE

    def __init__(self, by, reducer, conversion=None):
        super().__init__()
        self.by = [self.ensure_conversion(by_) for by_ in by]
        self.reducer = self.ensure_conversion(reducer)
        self.contents = self.contents & ~self.ContentTypes.REDUCER
        self.number_of_input_uses = 1
        self.aggregate_mode = len(self.by) == 0

        if conversion:
            self.conversion = self.ensure_conversion(conversion)
        else:
            self.conversion = self.ensure_conversion(
                ListComp(
                    GeneratorItem(
                        self.AGG_RESULT_ITEM,
                        self.SIGNATURE,
                        self.AGG_DATA,
                    ),
                    _none,
                    _none,
                )
                if len(self.by)
                else self.AGG_RESULT_ITEM
            )

    to_iter = delegate_input_switching_method("to_iter")
    iter = delegate_input_switching_method("iter", True)
    iter_mut = delegate_input_switching_method("iter_mut", True)
    iter_windows = delegate_input_switching_method("iter_windows", True)
    filter = delegate_input_switching_method("filter")
    flatten = delegate_input_switching_method("flatten", True)
    take_while = delegate_input_switching_method("take_while", True)
    drop_while = delegate_input_switching_method("drop_while", True)
    as_type = delegate_input_switching_method("as_type", True)
    sort = delegate_input_switching_method("sort", True)
    tap = delegate_input_switching_method("tap", True)

    def _gen_agg_data_container(self, container_name, number_of_reducers, ctx):
        attrs = [f"v{i}" for i in range(number_of_reducers)]

        code = Code()
        code.add_line(f"class {container_name}:", 1)
        _ = ",".join(f"'{attr}'" for attr in attrs)
        code.add_line(f"__slots__ = [{_}]", 0)
        code.add_line("def __init__(self, _none=__none__):", 1)
        if attrs:
            for attr in attrs:
                code.add_line(f"self.{attr} = _none", 0)
        else:
            code.add_line("pass", 0)
        return ctx[
            self.compile_converter(container_name, code.to_string(0), ctx)
        ]

    def _gen_code_and_update_ctx(self, code_input, ctx) -> str:
        ctx["defaultdict"] = defaultdict
        ctx["ListSortedOnceWrapper"] = ListSortedOnceWrapper

        suffix = self.gen_random_name("_", ctx)
        var_row = f"row{suffix}"
        var_signature = f"signature{suffix}"
        var_signature_to_agg_data = f"signature_to_agg_data{suffix}"
        var_agg_data = f"agg_data{suffix}"
        var_agg_data_cls = f"AggData{suffix}"

        function_ctx = self.as_function_ctx(ctx, optimize_naive=True)
        function_ctx.add_arg("data_", This())

        with function_ctx:
            reduce_blocks = ReduceBlocks(var_agg_data, self.aggregate_mode)
            if "exchange_reducer_code_input_methods" not in ctx:
                ctx["exchange_reducer_code_input_methods"] = []

            ctx["exchange_reducer_code_input_methods"].append(
                reduce_blocks.exchange_reducer_code_input
            )
            try:
                code_agg_result = self.reducer.gen_code_and_update_ctx(
                    var_row, ctx
                )
            finally:
                ctx["exchange_reducer_code_input_methods"].pop()
                if not ctx["exchange_reducer_code_input_methods"]:
                    del ctx["exchange_reducer_code_input_methods"]

            by_is_single = len(self.by) == 1
            code_signature = []
            for index, by_ in enumerate(self.by):
                code_by = by_.gen_code_and_update_ctx(var_row, ctx)
                code_signature.append(code_by)
                code_agg_result = self.replace_word(
                    code_agg_result,
                    code_by,
                    (
                        var_signature
                        if by_is_single
                        else f"{var_signature}[{index}]"
                    ),
                )

            code_signature = (
                code_signature[0]
                if by_is_single
                else f"({', '.join(code_signature)})"
            )

            if var_row in code_agg_result:
                raise ConversionException(
                    "something other than group_by keys and reducers have been used",
                    code_agg_result,
                )

            with NamespaceCtx(
                {
                    self.SIGNATURE_NAME: var_signature,
                    self.AGG_DATA_NAME: var_agg_data,
                    self.AGG_RESULT_ITEM_NAME: code_agg_result,
                },
                ctx,
            ):
                code_final_result = (
                    self.conversion.gen_code_and_update_ctx(None, ctx)
                    if self.aggregate_mode
                    else self.conversion.gen_code_and_update_ctx(
                        f"{var_signature_to_agg_data}.items()", ctx
                    )
                )

            agg_template_kwargs = dict(
                code_args=function_ctx.get_def_all_args_code(),
                code_result=f"    return {code_final_result}",
                var_row=var_row,
            )

            if self.aggregate_mode:
                converter_name = f"aggregate{suffix}"
                grouper_code = AGGREGATE_TEMPLATE.format(
                    converter_name=converter_name,
                    code_init_agg_vars=f"{' = '.join(reduce_blocks.agg_data_values)} = _none",
                    var_checksum=ReduceBlock.var_checksum,
                    code_aggregate=reduce_blocks.gen_aggregate_code(
                        var_row=var_row,
                        expected_checksum=reduce_blocks.number,
                    ).to_string(
                        base_indent_level=1,
                    ),
                    **agg_template_kwargs,
                )
            else:
                converter_name = f"group_by{suffix}"
                ctx[var_agg_data_cls] = self._gen_agg_data_container(
                    var_agg_data_cls, reduce_blocks.number, ctx
                )
                grouper_code = GROUPER_TEMPLATE.format(
                    converter_name=converter_name,
                    var_signature_to_agg_data=var_signature_to_agg_data,
                    var_agg_data_cls=var_agg_data_cls,
                    var_agg_data=var_agg_data,
                    code_signature=code_signature,
                    code_group_by=reduce_blocks.gen_group_by_code(
                        var_row=var_row,
                        var_agg_data=var_agg_data,
                        var_signature_to_agg_data=var_signature_to_agg_data,
                        code_signature=code_signature,
                    ).to_string(base_indent_level=1),
                    **agg_template_kwargs,
                )

            conversion = function_ctx.gen_conversion(
                converter_name, grouper_code
            )
        return function_ctx.call_with_all_args(
            conversion
        ).gen_code_and_update_ctx(code_input, ctx)


def Aggregate(  # pylint:disable=invalid-name
    *args, **kwargs
) -> BaseConversion:
    """Shortcut for ``GroupBy().aggregate(*args, **kwargs)``"""
    return GroupBy().aggregate(*args, **kwargs)


class ReducerDispatcher:
    pass


class SumReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = ({0} or 0)",)
    reduce = ("%(result)s = {prev_result} + ({0} or 0)",)
    default = NaiveConversion(0)
    unconditional_init = True


class FastSumReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = ({0} or 0)",)
    reduce = ("%(result)s = {prev_result} + {0}",)
    default = NaiveConversion(0)
    unconditional_init = True


class SumReducerDispatcher(ReducerDispatcher):
    def __call__(self, expression, *args, **kwargs):
        expression = ensure_conversion(expression)
        if expression.output_hints & expression.OutputHints.NOT_NONE:
            return FastSumReducer(expression, *args, **kwargs)
        return SumReducer(expression, *args, **kwargs)


class SumOrNoneReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = {0}",)
    reduce = (
        "if {0} is None:",
        "    %(result)s = None",
        "elif %(result)s is not None:",
        "    %(result)s = {prev_result} + {0}",
    )
    default = NaiveConversion(None)
    unconditional_init = True


class MaxReducer(MultiStatementReducer):
    base_condition_code = "{0} is not None"
    prepare_first = ("%(result)s = {0}",)
    reduce = (
        "if {0} > %(result)s:",
        "    %(result)s = {0}",
    )
    default = NaiveConversion(None)
    unconditional_init = True


class MinReducer(MultiStatementReducer):
    base_condition_code = "{0} is not None"
    prepare_first = ("%(result)s = {0}",)
    reduce = (
        "if {0} < %(result)s:",
        "    %(result)s = {0}",
    )
    default = NaiveConversion(None)
    unconditional_init = True


class CountReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = 1",)
    reduce = ("%(result)s = {prev_result} + 1",)
    default = NaiveConversion(0)
    unconditional_init = True


class CountDistinctReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = {{ {0} }}",)
    reduce = ("%(result)s.add({0})",)
    default = NaiveConversion(0)
    unconditional_init = True
    post_conversion = CallFunc(len, This())


class FirstReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = {0}",)
    reduce = ()
    default = NaiveConversion(None)
    unconditional_init = True


class LastReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = {0}",)
    reduce = ("%(result)s = {0}",)
    default = NaiveConversion(None)
    unconditional_init = True


class MaxRowReducer(MultiStatementReducer):
    """Reducer which finds an item with max value of the expression and returns
    this item"""

    base_condition_code = "{0} is not None"
    prepare_first = ("%(result)s = ({0}, %(input)s)",)
    reduce = (
        "if %(result)s[0] < {0}:",
        "    %(result)s = ({0}, %(input)s)",
    )
    default = NaiveConversion(None)
    post_conversion = GetItem(1)
    unconditional_init = True


class MinRowReducer(MaxRowReducer):
    reduce = (
        "if %(result)s[0] > {0}:",
        "    %(result)s = ({0}, %(input)s)",
    )


class ArrayReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = [{0}]",)
    reduce = ("%(result)s.append({0})",)
    default = NaiveConversion(None)
    unconditional_init = True


class ListSortedOnceWrapper:
    """Wraps a list, exposes append method only. Once the list is filled up, it
    is sorted (only once) in-place and is returned when get method called."""

    __slots__ = ["list_", "append", "sorted", "key", "reverse"]

    def __init__(self, list_: list, key=None, reverse=False):
        self.list_ = list_
        self.append = self.list_.append
        self.sorted = False
        self.key = key
        self.reverse = reverse

    def get(self) -> list:
        if not self.sorted:
            self.list_.sort(key=self.key, reverse=self.reverse)
            self.sorted = True
            del self.append
        return self.list_


class SortedArrayReducer(MultiStatementReducer):
    """Array reducer which sorts the list just once in the end"""

    reduce = ("%(result)s.append({0})",)
    default = NaiveConversion(None)
    unconditional_init = True

    def __init__(self, *args, key=None, reverse=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.key = self.ensure_conversion(key)
        self.reverse = self.ensure_conversion(reverse)

    def prepare_first(self, ctx):
        key_code = self.key.gen_code_and_update_ctx("{0}", ctx)
        reverse_code = self.reverse.gen_code_and_update_ctx("{0}", ctx)
        return (
            "%(result)s = ListSortedOnceWrapper("
            f"[{{0}}], {key_code}, {reverse_code})",
        )

    @property
    def post_conversion(self):
        return This().call_method("get")


class ArrayDistinctReducer(MultiStatementReducer):
    prepare_first = ("%(result)s = {{ {0}: None }}",)
    reduce = ("%(result)s[{0}] = None",)
    post_conversion = InlineExpr("list({0})").pass_args(This())
    default = NaiveConversion(None)
    unconditional_init = True


class BaseDictReducer(MultiStatementReducer):
    """This reducer accepts 2 expressions:

    - the first one is used to calculate keys of the resulting dict
    - the second one is used to calculate values to be reduced and put as the
      final value, under the certain key.

    Effectively dict reducers allow for double grouping: the first one on the
    top level, the second one on DictReducer level.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # for backward compatibility with versions <= 0.8
        if len(self.expressions) == 1:
            expr = self.expressions[0]
            if isinstance(expr, (Tuple, List)) and len(expr.items) == 2:
                self.expressions = tuple(expr.items)
        if len(self.expressions) != 2:
            raise ValueError("invalid dict reducer input: two values expected")


class DictReducer(BaseDictReducer):
    prepare_first = ("%(result)s = {{ {0}: {1} }}",)
    reduce = ("%(result)s[{0}] = {1}",)
    default = NaiveConversion(None)
    unconditional_init = True


lock_default_dict_conversion = InlineExpr(
    'setattr({this_}, "default_factory", None) or {this_}'
).pass_args(this_=This())


class DictArrayReducer(BaseDictReducer):
    prepare_first = (
        "%(result)s = _d = defaultdict(list)",
        "_d[{0}].append({1})",
    )
    reduce = ("%(result)s[{0}].append({1})",)
    post_conversion = lock_default_dict_conversion
    default = NaiveConversion(None)
    unconditional_init = True


class DictArrayDistinctReducer(BaseDictReducer):
    """Dict reducer where dict values are lists of distinct values"""

    prepare_first = (
        "%(result)s = _d = defaultdict(dict)",
        "_d[{0}][{1}] = None",
    )
    reduce = ("%(result)s[{0}][{1}] = None",)
    post_conversion = InlineExpr(
        "{{k_: list(v_) for k_, v_ in {}.items()}}"
    ).pass_args(This())
    default = NaiveConversion(None)
    unconditional_init = True


class DictSumReducer(BaseDictReducer):
    prepare_first = (
        "%(result)s = _d = defaultdict(int)",
        "_d[{0}] = {1} or 0",
    )
    reduce = ("%(result)s[{0}] = {prev_result}[{0}] + ({1} or 0)",)
    post_conversion = lock_default_dict_conversion
    default = NaiveConversion(None)
    unconditional_init = True


class DictSumOrNoneReducer(BaseDictReducer):
    """Dict reducer where dict values are either numbers or None if there's
    been at least one None value within the group"""

    prepare_first = ("%(result)s = _d = defaultdict(int)", "_d[{0}] = {1}")
    reduce = (
        "if {1} is None:",
        "    %(result)s[{0}] = None",
        "elif %(result)s[{0}] is not None:",
        "    %(result)s[{0}] = {prev_result}[{0}] + {1}",
    )
    post_conversion = lock_default_dict_conversion
    default = NaiveConversion(None)
    unconditional_init = True


class DictMaxReducer(BaseDictReducer):
    """DictMax reducer which takes first positional item as keys and
    accumulates max value of second positional item"""

    base_condition_code = "{1} is not None"
    prepare_first = ("%(result)s = {{ {0}: {1} }}",)
    reduce = (
        "if {0} not in %(result)s or {1} > %(result)s[{0}]:",
        "    %(result)s[{0}] = {1}",
    )
    default = NaiveConversion(None)
    unconditional_init = True


class DictMinReducer(DictMaxReducer):
    reduce = (
        "if {0} not in %(result)s or {1} < %(result)s[{0}]:",
        "    %(result)s[{0}] = {1}",
    )


class DictCountReducer(BaseDictReducer):
    prepare_first = ("%(result)s = {{ {0}: 1 }}",)
    reduce = (
        "if {0} not in %(result)s:",
        "    %(result)s[{0}] = 1",
        "else:",
        "    %(result)s[{0}] = {prev_result}[{0}] + 1",
    )
    default = NaiveConversion(None)
    unconditional_init = True


class DictCountDistinctReducer(BaseDictReducer):
    """Dict reducer where dict values are numbers of distinct values per
    group"""

    prepare_first = ("%(result)s = {{ {0}: {{ {1} }} }}",)
    reduce = (
        "if {0} not in %(result)s:",
        "    %(result)s[{0}] = {{ {1} }}",
        "else:",
        "    %(result)s[{0}].add({1})",
    )
    post_conversion = InlineExpr(
        "{{ k_: len(v_) for k_, v_ in {}.items() }}"
    ).pass_args(This())
    default = NaiveConversion(None)
    unconditional_init = True


class DictFirstReducer(BaseDictReducer):
    prepare_first = ("%(result)s = {{ {0}: {1} }}",)
    reduce = ("if {0} not in %(result)s:", "    %(result)s[{0}] = {1}")
    default = NaiveConversion(None)
    unconditional_init = True


class DictLastReducer(BaseDictReducer):
    prepare_first = ("%(result)s = {{ {0}: {1} }}",)
    reduce = ("%(result)s[{0}] = {1}",)
    default = NaiveConversion(None)
    unconditional_init = True


class AverageReducerDispatcher(ReducerDispatcher):
    """Dispatcher which chooses between weighted and simple averages"""

    sum_reducer_dispatcher = SumReducerDispatcher()

    def __call__(
        self, value, weight=1, default=None, where=None
    ) -> "BaseConversion":
        """
        Calculates the arithmetic mean or weighted mean.
        """
        if isinstance(weight, (int, float, Decimal)) and weight == 1:
            return If(
                CountReducer(where=where),
                (
                    self.sum_reducer_dispatcher(value, where=where)
                    / CountReducer(where=where)
                ),
                default,
            )
        return If(
            self.sum_reducer_dispatcher(weight, where=where),
            (
                self.sum_reducer_dispatcher(value * weight, where=where)
                / self.sum_reducer_dispatcher(weight, where=where)
            ),
            default,
        )


class TopReducer(DictCountReducer):
    """
    Returns a list of the most frequent values.
    The resulting list is sorted in descending order of values frequency.
    """

    def __init__(self, k: int, key_conv, *args, **kwargs):
        if not isinstance(k, int):
            raise TypeError("K must be an integer.")

        if k < 1:
            raise ValueError("K must be a positive integer greater than 0.")

        self.k = k
        super().__init__(key_conv, 1, *args, **kwargs)

    @property
    def post_conversion(self):
        return InlineExpr(
            "[k for k,v in sorted((v,k) for k,v in {data}.items())[:-{k}:-1]]"
        ).pass_args(data=This(), k=self.k + 1)


class ModeReducer(DictCountReducer):
    def __init__(self, conv, *args, **kwargs):
        super().__init__(conv, conv, *args, **kwargs)

    post_conversion = InlineExpr(
        "sorted(((v,k) for k,v in {data}.items()))[-1][1]"
    ).pass_args(data=This())


class PercentileReducer(SortedArrayReducer):
    """Calculates percentile (floats from 0 to 100 inclusive)

    >>> c.ReduceFuncs.Percentile(95, c.item("amount"))
    >>> c.ReduceFuncs.Percentile(95, c.item("amount"), interpolation="lower")

    interpolation options:
      * "linear"
      * "lower"
      * "higher"
      * "midpoint"
      * "nearest"
    """

    interpolation_to_method: "t.Dict[str, t.Callable]" = {}

    def __init__(
        self, percentile: float, conv, *args, interpolation="linear", **kwargs
    ):
        """

        Args:
         * interpolation: one of
           #. "linear"
           #. "lower"
           #. "higher"
           #. "midpoint"
           #. "nearest"
        """

        self.percentile = percentile
        try:
            self.method = self.interpolation_to_method[interpolation]
        except KeyError as e:
            raise ValueError("unsupported interpolation type") from e

        super().__init__(conv, *args, **kwargs)
        if not 0 <= percentile <= 100:
            raise ValueError(
                "percentile must be a float between 0 and 100 inclusive"
            )

    @staticmethod
    def percentile_linear(data, quantile):
        max_index = len(data) - 1
        index = max_index * quantile
        left_index = int(index)
        left_value = data[left_index]
        if left_index == max_index:
            return left_value

        return left_value + (data[left_index + 1] - left_value) * (
            index - left_index
        )

    @staticmethod
    def percentile_lower(data, quantile):
        return data[int((len(data) - 1) * quantile)]

    @staticmethod
    def percentile_higher(data, quantile):
        return data[ceil((len(data) - 1) * quantile)]

    @staticmethod
    def percentile_midpoint(data, quantile):
        index = (len(data) - 1) * quantile
        left_index = int(index)
        if left_index == index:
            return data[left_index]

        left_value = data[left_index]
        return left_value + (data[left_index + 1] - left_value) * 0.5

    @staticmethod
    def percentile_nearest(data, quantile):
        index = (len(data) - 1) * quantile
        left_index = int(index)
        if index - left_index > 0.5:
            return data[left_index + 1]
        else:
            return data[left_index]

    @property
    def post_conversion(self):
        return CallFunc(
            self.method, super().post_conversion, self.percentile * 0.01
        )


PercentileReducer.interpolation_to_method = {
    "linear": PercentileReducer.percentile_linear,
    "lower": PercentileReducer.percentile_lower,
    "higher": PercentileReducer.percentile_higher,
    "midpoint": PercentileReducer.percentile_midpoint,
    "nearest": PercentileReducer.percentile_nearest,
}


def MedianReducer(  # pylint:disable=invalid-name
    conv, *args, **kwargs
) -> BaseConversion:
    return PercentileReducer(50, conv, *args, **kwargs)


class ReduceFuncs:
    """Exposes the list of reduce functions"""

    # pylint: disable=invalid-name

    #: Calculates the sum, skips false values
    Sum = SumReducerDispatcher()
    #: Calculates the sum, any ``None`` makes the total sum ``None``
    SumOrNone = SumOrNoneReducer

    #: Finds max value, skips ``None``
    Max = MaxReducer
    #: Finds a row with max value, skips ``None``
    MaxRow = MaxRowReducer

    #: Finds min value, skips ``None``
    Min = MinReducer
    #: Finds a row with min value, skips ``None``
    MinRow = MinRowReducer

    #: Counts objects
    Count = CountReducer
    #: Counts distinct values
    CountDistinct = CountDistinctReducer

    #: Stores the first value per group
    First = FirstReducer
    #: Stores the last value per group
    Last = LastReducer

    #: Calculates the arithmetic mean or weighted mean.
    Average = AverageReducerDispatcher()
    #: Calculates the median value.
    Median = MedianReducer
    #: Calculates percentile: floats in [0, 100]
    Percentile = PercentileReducer
    #: Calculates the most common value.
    #: In case of multiple values, returns the last of them.
    Mode = ModeReducer
    #: Returns a list of the most frequent values.
    #: The resulting list is sorted in descending order of values frequency.
    TopK = TopReducer

    #: Aggregates values into array
    Array = ArrayReducer
    #: Aggregates distinct values into array, preserves order
    ArrayDistinct = ArrayDistinctReducer
    #: Aggregates values into array, sorting them in the end
    ArraySorted = SortedArrayReducer

    #: Aggregates values into dict; dict values are last values per group
    Dict = DictReducer
    #: Aggregates values into dict; dict values are lists of group values
    DictArray = DictArrayReducer
    #: Aggregates values into dict; dict values are lists of unique group
    #: values preserves order
    DictArrayDistinct = DictArrayDistinctReducer
    #: Aggregates values into dict; dict values are sums of group values,
    #: skipping ``None``
    DictSum = DictSumReducer
    #: Aggregates values into dict; dict values are sums of group values,
    #: any ``None`` makes the total sum ``None``
    DictSumOrNone = DictSumOrNoneReducer
    #: Aggregates values into dict; dict values are max group values
    DictMax = DictMaxReducer
    #: Aggregates values into dict; dict values are min group values
    DictMin = DictMinReducer
    #: Aggregates values into dict; dict values are numbers of values in groups
    DictCount = DictCountReducer
    #: Aggregates values into dict; dict values are numbers of unique values
    #: in groups
    DictCountDistinct = DictCountDistinctReducer
    #: Aggregates values into dict; dict values are first values per group
    DictFirst = DictFirstReducer
    #: Aggregates values into dict; dict values are last values per group
    DictLast = DictLastReducer



"""Helpers live here: like:
 - recently used cache
 - options ctx manager
"""
import os
import sys
import tempfile
import threading
import typing as t
from collections import deque
from importlib import import_module
from typing import TYPE_CHECKING
from weakref import finalize


if TYPE_CHECKING:
    from typing import Optional

PY_VERSION = sys.version_info[:2]
if PY_VERSION == (3, 6):

    class BaseCtxMeta(t.GenericMeta):  # type: ignore
        def __init__(
            cls, name, bases, kwargs
        ):  # pylint: disable=no-self-argument
            super().__init__(name, bases, kwargs)
            cls._ctx = threading.local()

else:

    class BaseCtxMeta(type):  # type: ignore
        def __init__(cls, name, bases, kwargs):
            super().__init__(name, bases, kwargs)
            cls._ctx = threading.local()


class BaseOptionsMeta(type):
    def __init__(cls, name, bases, kwargs):
        super().__init__(name, bases, kwargs)
        cls._option_attrs = {
            k: v
            for k, v in kwargs.items()
            if not k.startswith("_") and k not in {"clone", "to_defaults"}
        }


class BaseOptions(object, metaclass=BaseOptionsMeta):
    """Container object, which carries current options"""

    def clone(self):
        clone = self.__class__()
        for option_attr in self._option_attrs.keys():
            setattr(clone, option_attr, getattr(self, option_attr))
        return clone

    def to_defaults(self, option_name=None):
        if option_name:
            setattr(self, option_name, self._option_attrs[option_name])
        else:
            for option_attr, value in self._option_attrs.items():
                setattr(self, option_attr, value)


OT = t.TypeVar("OT", bound=BaseOptions)


class BaseCtx(
    t.Generic[OT], metaclass=BaseCtxMeta
):  # pylint:disable=invalid-metaclass
    """Context manager to manage option objects"""

    options_cls: t.Type[OT]
    _ctx: threading.local

    def __enter__(self) -> OT:
        self._ctx.prev_options = getattr(self._ctx, "options", None)
        if self._ctx.prev_options:
            self._ctx.options = self._ctx.prev_options.clone()
        else:
            self._ctx.options = self.options_cls()
        return self._ctx.options

    def __exit__(self, exc_type, exc_value, tb):
        self._ctx.options = self._ctx.prev_options
        self._ctx.prev_options = None

    @classmethod
    def get_option_value(cls, option_name):
        options = getattr(cls._ctx, "options", None)
        if not options:
            options = cls.options_cls
        return getattr(options, option_name)


class Code:
    """A building block for generating code, which is composed of multiple
    statements."""

    def __init__(self):
        self.lines_info = []
        self.indent_level = 0
        self.expression_line = None

    def add_line(
        self,
        line: str,
        next_line_indent_incr: int,
        expression_line: "Optional[str]" = None,
    ):
        self.lines_info.append((self.indent_level, line))
        self.indent_level += next_line_indent_incr
        self.expression_line = expression_line
        if self.indent_level < 0:
            raise AssertionError("negative indentation level")

    def as_expression(self) -> "Optional[str]":
        if len(self.lines_info) == 1 and self.expression_line:
            return self.expression_line
        return None

    def add_code(self, code: "Code"):
        for indent_level, line in code.lines_info:
            self.lines_info.append((indent_level + self.indent_level, line))

    def incr_indent_level(self, incr: int):
        self.indent_level += incr

    def has_lines(self):
        return bool(self.lines_info)

    def to_string(self, base_indent_level: int, single_indent: str = "    "):
        return "\n".join(
            f"{single_indent * (base_indent_level + indent_level)}{line}"
            for indent_level, line in self.lines_info
        )


class LazyDebugDir:
    """Lazy instance to hold and initialize debug directory for generated code
    sources"""

    def __init__(self):
        self.debug_dir = None
        self.dir_initialized = False

    def get(self):
        if self.debug_dir is None:
            self.debug_dir = os.environ.get(
                "PY_CONVTOOLS_DEBUG_DIR", None
            ) or os.path.join(tempfile.gettempdir(), "py_convtools_debug")
        return self.debug_dir

    def ensure_initialized(self):
        if not self.dir_initialized:
            os.makedirs(self.debug_dir, exist_ok=True)
            self.dir_initialized = True


debug_dir = LazyDebugDir()


class CodePiece:
    """Piece of source code (a function at the moment)"""

    __slots__ = (
        "converter_name",
        "code_parts",
        "abs_path",
        "is_dumped",
    )

    def __init__(self, converter_name, code_parts, abs_path, is_dumped):
        self.converter_name = converter_name
        self.code_parts = code_parts
        self.abs_path = abs_path
        self.is_dumped = is_dumped


class CodeStorage:
    """Container which stores generated code pieces. It allows to dump code
    sources on disk into a debug directory."""

    def __init__(self):
        self.key_to_code_piece = {}
        self.converter_names = set()
        finalize(self, drop_dumped_code, self.key_to_code_piece)

    def add_sources(self, converter_name, code_str):
        def_name = f"def {converter_name}("
        code_parts = (def_name, code_str.replace(def_name, ""))

        code_piece = self.key_to_code_piece.get(code_parts[1])
        if code_piece is not None:
            return code_piece, False

        if converter_name in self.converter_names:
            raise Exception(
                "converter with a different code already exists",
                converter_name,
            )
        self.converter_names.add(converter_name)

        abs_path = os.path.join(
            debug_dir.get(), f"_{id(self)}_{converter_name}.py"
        )
        code_piece = self.key_to_code_piece[code_parts[1]] = CodePiece(
            converter_name, code_parts, abs_path, False
        )
        return code_piece, True

    def dump_sources(self):
        debug_dir.ensure_initialized()
        for code_piece in self.key_to_code_piece.values():
            if not code_piece.is_dumped:
                with open(code_piece.abs_path, "w", encoding="utf-8") as f:
                    f.write("".join(code_piece.code_parts))
                code_piece.is_dumped = True


def drop_dumped_code(key_to_code_piece):
    for code_piece in key_to_code_piece.values():
        if code_piece.is_dumped:
            try:
                os.remove(code_piece.abs_path)
            except FileNotFoundError:  # pragma: no cover
                pass


def iter_windows(collection, width, step):
    window = deque(maxlen=width)
    window_append = window.append

    index = 0
    for index, obj in enumerate(collection):
        window_append(obj)
        if index % step == 0:
            yield tuple(window)

    if window:
        index += 1
        window.popleft()
        while window:
            if index % step == 0:
                yield tuple(window)
            index += 1
            window.popleft()


obj_getattribute = object.__getattribute__


class LazyModule:
    """Lazy import helper, which caches results of importlib.import_module"""

    __slots__ = ["name", "package", "_module"]

    def __init__(self, name, package=None):
        self.name = name
        self.package = package
        self._module = None

    def __getattribute__(self, name):
        module = obj_getattribute(self, "_module")
        if module is None:
            module = self._module = import_module(
                obj_getattribute(self, "name"),
                obj_getattribute(self, "package"),
            )
        return getattr(module, name)



"""Provides cumulative conversions"""
import typing as t
from uuid import uuid4

from .base import (
    BaseConversion,
    EscapedString,
    If,
    LabelConversion,
    LazyEscapedString,
    NaiveConversion,
    Namespace,
    PipeConversion,
)


class CumulativeReset(BaseConversion):
    """A conversion which resets cumulative values to their initial states. The
    main use case is within nested iterables:

    >>> assert (
    >>>     c.iter(
    >>>         c.cumulative_reset("abc")
    >>>         .iter(c.cumulative(c.this, c.this + c.PREV, label_name="abc"))
    >>>         .as_type(list)
    >>>     )
    >>>     .as_type(list)
    >>>     .execute([[0, 1, 2], [3, 4]])
    >>> ) == [[0, 1, 3], [3, 7]]
    """

    def __init__(self, parent: "t.Any", label_name: str):
        super().__init__()
        self.label_name = label_name
        self.parent = self.ensure_conversion(parent)
        self.contents |= BaseConversion.ContentTypes.NEW_LABEL

    def _gen_code_and_update_ctx(self, code_input, ctx):
        return (
            f"(_labels.pop({repr(self.label_name)}, None), "
            f"{self.parent.gen_code_and_update_ctx(code_input, ctx)})[1]"
        )


class Cumulative(BaseConversion):
    """A conversion which calculates cumulative values. The main use case is
    using it within iterables:

    >>> assert (
    >>>     c.iter(c.cumulative(c.this, c.this + c.PREV))
    >>>     .as_type(list)
    >>>     .execute([0, 1, 2, 3, 4])
    >>> ) == [0, 1, 3, 6, 10]
    """

    PREV = LazyEscapedString("prev_value")

    def __init__(
        self,
        parent: "t.Any",
        prepare_first: "t.Any",
        reduce_two: "t.Any",
        label_name: "t.Optional[str]" = None,
    ):
        """
        Args:
          parent: conversion which is used as an input
          prepare_first: conversion which gets initial value from the first
            element
          reduce_two: conversion which reduces two values to one
          label_name: custom name of cumulative to be used. It is needed when
            :py:obj:`CumulativeReset`
        """
        super().__init__()
        self.label_name = label_name or uuid4().hex

        self.parent = self.ensure_conversion(parent)

        label_conversion = LabelConversion(self.label_name)

        self.prepare_first = self.ensure_conversion(prepare_first)
        self.reduce_two = self.ensure_conversion(
            Namespace(
                reduce_two,
                name_to_code={
                    self.PREV.name: label_conversion.gen_code_and_update_ctx(
                        None, label_conversion._init_ctx()
                    )
                },
            )
        )
        self.contents |= BaseConversion.ContentTypes.NEW_LABEL

    def _gen_code_and_update_ctx(self, code_input, ctx):
        return PipeConversion(
            self.parent,
            If(
                NaiveConversion(self.label_name).in_(
                    EscapedString(LabelConversion.labels_code_name)
                ),
                self.reduce_two,
                self.prepare_first,
            ),
            label_output=self.label_name,
        ).gen_code_and_update_ctx(code_input, ctx)



"""
This module is to contain helpers, which collect info about python code
execution
"""
import sys


# generated by "python benchmarks/build_heuristics.py"

PY_VERSION = sys.version_info[0:2]
if PY_VERSION <= (3, 6):

    # 3.6.15
    class Weights:  # type: ignore # pragma: no cover # pylint: disable=missing-class-docstring # noqa: F811
        # base_time: 4.0568550033715e-11
        STEP = 100
        LOGICAL = 251
        DICT_LOOKUP = 327
        MATH_SIMPLE = 377
        ATTR_LOOKUP = 342
        TUPLE_INIT = 538
        LIST_INIT = 675
        SET_INIT = 1711
        DICT_INIT = 1667
        FUNCTION_CALL = DICT_LOOKUP * 4
        UNPREDICTABLE = 171100

elif PY_VERSION <= (3, 7):  # pragma: no cover

    # 3.7.16
    class Weights:  # type: ignore # pragma: no cover # pylint: disable=missing-class-docstring # noqa: F811
        # base_time: 4.08321829275715e-11
        STEP = 100
        LOGICAL = 236
        DICT_LOOKUP = 337
        MATH_SIMPLE = 365
        ATTR_LOOKUP = 319
        TUPLE_INIT = 525
        LIST_INIT = 655
        SET_INIT = 1643
        DICT_INIT = 1685
        FUNCTION_CALL = DICT_LOOKUP * 4
        UNPREDICTABLE = 168500

elif PY_VERSION == (3, 8):  # pragma: no cover

    # 3.8.16
    class Weights:  # type: ignore # pragma: no cover # pylint: disable=missing-class-docstring # noqa: F811
        # base_time: 4.144752377655609e-11
        STEP = 100
        LOGICAL = 226
        DICT_LOOKUP = 319
        MATH_SIMPLE = 368
        ATTR_LOOKUP = 309
        TUPLE_INIT = 532
        LIST_INIT = 675
        SET_INIT = 1677
        DICT_INIT = 1758
        FUNCTION_CALL = DICT_LOOKUP * 4
        UNPREDICTABLE = 175800

elif PY_VERSION == (3, 9):  # pragma: no cover

    # 3.9.12
    class Weights:  # type: ignore # pragma: no cover # pylint: disable=missing-class-docstring # noqa: F811
        # base_time: 4.01792780282291e-11
        STEP = 100
        LOGICAL = 241
        DICT_LOOKUP = 345
        MATH_SIMPLE = 409
        ATTR_LOOKUP = 328
        TUPLE_INIT = 610
        LIST_INIT = 775
        SET_INIT = 1422
        DICT_INIT = 1912
        FUNCTION_CALL = DICT_LOOKUP * 4
        UNPREDICTABLE = 191200

elif PY_VERSION == (3, 10):  # pragma: no cover

    # 3.10.9
    class Weights:  # type: ignore # pragma: no cover # pylint: disable=missing-class-docstring # noqa: F811
        # base_time: 4.1176799481692785e-11
        STEP = 100
        LOGICAL = 206
        DICT_LOOKUP = 306
        MATH_SIMPLE = 365
        ATTR_LOOKUP = 301
        TUPLE_INIT = 565
        LIST_INIT = 668
        SET_INIT = 1431
        DICT_INIT = 1798
        FUNCTION_CALL = DICT_LOOKUP * 4
        UNPREDICTABLE = 179800

else:

    # 3.11.1
    class Weights:  # type: ignore # pragma: no cover # pylint: disable=missing-class-docstring # noqa: F811
        # base_time: 4.285590663856272e-11
        STEP = 100
        LOGICAL = 178
        DICT_LOOKUP = 313
        MATH_SIMPLE = 343
        ATTR_LOOKUP = 282
        TUPLE_INIT = 530
        LIST_INIT = 647
        SET_INIT = 1379
        DICT_INIT = 1958
        FUNCTION_CALL = DICT_LOOKUP * 4
        UNPREDICTABLE = 195800



"""
Provides conversions for slicing iterables into chunks.
"""
import typing as t

from .aggregations import Aggregate
from .base import BaseConversion, Code, LazyEscapedString, Namespace, This


_none = BaseConversion._none


class BaseChunkBy(BaseConversion):
    def aggregate(self, *args, **kwargs) -> "BaseConversion":
        return self.iter(Aggregate(*args, **kwargs))


class ChunkBy(BaseChunkBy):
    """Slices iterable into chunks by element values and/or size of chunks.

    >>> # simple #1
    >>> c.chunk_by(size=1000)
    >>>
    >>> # simple #2
    >>> c.chunk_by(c.item("x"))
    >>>
    >>> # with aggregate
    >>> c.chunk_by(
    >>>     c.item("x"),
    >>>     size=1000
    >>> ).aggregate({
    >>>     "x": c.ReduceFuncs.Last(c.item("x")),
    >>>     "y": c.ReduceFuncs.Sum(c.item("y")),
    >>> })


    >>> # example #1
    >>> c.chunk_by_condition(c.CHUNK.len() < 100)
    >>>
    >>> # example #2
    >>> c.chunk_by_condition(
    >>>     c.and_(
    >>>         c.CHUNK.item(-1) == c.this,
    >>>         c.CHUNK.item(-1) - c.this < 100
    >>>     )
    >>> )
    >>> # with aggregate
    >>> c.chunk_by_condition(
    >>>     c.CHUNK.len() < 100
    >>> ).aggregate({
    >>>     "x": c.ReduceFuncs.Last(c.item("x")),
    >>>     "y": c.ReduceFuncs.Sum(c.item("y")),
    >>> })

    It also provides a shortcut for running
    :py:obj:`convtools.aggregations.Aggregate` on chunks.
    """

    def __init__(self, *by, size: t.Optional[int] = None):
        """
        Args:
          by: fields/conversions to use for slicing into chunks (elements with
            same values go to the same chunk)
          size: (optional) positive int to limit max size of a chunk
        """
        super().__init__()
        if size is not None:
            if not isinstance(size, int) or size <= 0:
                raise ValueError("size has to be positive int or None")
        if not by and not size:
            raise ValueError("pass at least one of by or size params")

        self.by = (
            (self.ensure_conversion(by if len(by) > 1 else by[0]))
            if by
            else None
        )
        self.size = size

    def _gen_code_and_update_ctx(self, code_input, ctx):
        converter_name = self.gen_random_name("chunk_by", ctx)
        function_ctx = (self.by or This()).as_function_ctx(
            ctx, optimize_naive=True
        )
        function_ctx.add_arg("items_", This())
        with function_ctx:
            code = Code()
            code.add_line("def placeholder", 1)
            code.add_line("items_ = iter(items_)", 0)
            code.add_line("try:", 0)
            code.add_line("    item_ = next(items_)", 0)
            code.add_line("except StopIteration:", 0)
            code.add_line("    return", 0)

            code.add_line("chunk_ = [item_]", 0)

            code_item_to_signature = (
                self.by.gen_code_and_update_ctx("item_", ctx)
                if self.by
                else None
            )

            code_before_for = Code()
            code_after_for = Code()
            code_if_condition = None
            code_if_continue_chunk = Code()
            code_if_new_chunk = Code()

            code_if_continue_chunk.add_line("chunk_.append(item_)", 0)
            code_if_new_chunk.add_line("yield chunk_", 0)
            code_if_new_chunk.add_line("chunk_ = [item_]", 0)

            if code_item_to_signature and self.size:
                code_if_condition = (
                    "chunk_item_signature == new_item_signature and size_ < "
                    f"{self.size}"
                )

            if code_item_to_signature:
                code_before_for.add_line(
                    f"chunk_item_signature = {code_item_to_signature}", 0
                )
                code_after_for.add_line(
                    f"new_item_signature = {code_item_to_signature}", 0
                )
                code_if_condition = (
                    code_if_condition
                    or "chunk_item_signature == new_item_signature"
                )
                code_if_new_chunk.add_line(
                    "chunk_item_signature = new_item_signature", 0
                )

            if self.size:
                code_before_for.add_line("size_ = 1", 0)
                code_if_condition = code_if_condition or f"size_ < {self.size}"
                code_if_continue_chunk.add_line("size_ = size_ + 1", 0)
                code_if_new_chunk.add_line("size_ = 1", 0)

            if not code_if_condition:
                raise AssertionError("impossible case")

            code.add_code(code_before_for)
            code.add_line("for item_ in items_:", 1)
            code.add_code(code_after_for)
            code.add_line(f"if {code_if_condition}:", 1)
            code.add_code(code_if_continue_chunk)
            code.incr_indent_level(-1)
            code.add_line("else:", 1)
            code.add_code(code_if_new_chunk)
            code.incr_indent_level(-2)
            code.add_line("yield chunk_", -1)

            code.lines_info[0] = (
                0,
                f"def {converter_name}({function_ctx.get_def_all_args_code()}):",
            )
            conversion = function_ctx.gen_conversion(
                converter_name, code.to_string(0)
            )

        return function_ctx.call_with_all_args(
            conversion
        ).gen_code_and_update_ctx(code_input, ctx)


CHUNK_BY_CONDITION_TEMPLATE = """
def {converter_name}({code_args}):
    items_ = iter(items_)
    try:
        chunk_ = [next(items_)]
    except StopIteration:
        return

    for item_ in items_:
        if {code_condition}:
            chunk_.append(item_)
        else:
            yield chunk_
            chunk_ = [item_]

    yield chunk_
"""


class ChunkByCondition(BaseChunkBy):
    """Slices iterable into chunks based on condition, which is a function of a
    current chunk and a current element.

    >>> # example #1
    >>> c.chunk_by_condition(c.CHUNK.len() < 100)
    >>>
    >>> # example #2
    >>> c.chunk_by_condition(
    >>>     c.and_(
    >>>         c.CHUNK.item(-1) == c.this,
    >>>         c.CHUNK.item(-1) - c.this < 100
    >>>     )
    >>> )
    >>> # with aggregate
    >>> c.chunk_by_condition(
    >>>     c.CHUNK.len() < 100
    >>> ).aggregate({
    >>>     "x": c.ReduceFuncs.Last(c.item("x")),
    >>>     "y": c.ReduceFuncs.Sum(c.item("y")),
    >>> })

    It also provides a shortcut for running
    :py:obj:`convtools.aggregations.Aggregate` on chunks.
    """

    CHUNK = LazyEscapedString("chunk_")

    def __init__(self, condition):
        super().__init__()
        self.condition = self.ensure_conversion(
            Namespace(condition, {self.CHUNK.name: "chunk_"})
        )

    def _gen_code_and_update_ctx(self, code_input, ctx):
        converter_name = self.gen_random_name("chunk_by_condition", ctx)
        function_ctx = self.condition.as_function_ctx(ctx, optimize_naive=True)
        function_ctx.add_arg("items_", This())

        with function_ctx:
            code_condition = self.condition.gen_code_and_update_ctx(
                "item_", ctx
            )
            code = CHUNK_BY_CONDITION_TEMPLATE.format(
                converter_name=converter_name,
                code_args=function_ctx.get_def_all_args_code(),
                code_condition=code_condition,
            )
            conversion = function_ctx.gen_conversion(converter_name, code)
        return function_ctx.call_with_all_args(
            conversion
        ).gen_code_and_update_ctx(code_input, ctx)
